//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
// Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009 James Lehman.
// This source is distributed under the terms of the GNU General Public License.
//
// LaserBoy_TUI.cpp is part of LaserBoy.
//
// LaserBoy is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// LaserBoy is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LaserBoy. If not, see <http://www.gnu.org/licenses/>.
//
//############################################################################
#include "LaserBoy_GUI_base.hpp"

//############################################################################
void LaserBoy_TUI::overlay_palettes()
{
    int    i,
           j,
           y_line       = 8 * font_size_factor,
           y_off        =   LASERBOY_PALETTE_MAX * font_size_factor
                          + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    //------------------------------------------------------------------------
    if(palette_index != current_frame().palette_index)
        palette_index = current_frame().palette_index;
    //------------------------------------------------------------------------
    if(palette_index != LASERBOY_TRUE_COLOR) // 8 bit frame palette
    {
        for(i = 0; i < const_palette_picker(palette_index).number_of_colors(); i++)
            for(j = 0; j < font_size_factor; j++)
                bmp_put_line_fast( &bmp,
                                   x_off,
                                   i * font_size_factor + j,
                                   x_off + 50,
                                   i * font_size_factor + j,
                                   bmp.make_rgb(const_palette_picker(palette_index)[i].r,
                                                const_palette_picker(palette_index)[i].g,
                                                const_palette_picker(palette_index)[i].b
                                               )
                                 );
    }
    //------------------------------------------------------------------------
    for(i = 0; i < const_palette_picker(target_palette_index).number_of_colors(); i++) // target palette
        for(j = 0; j < font_size_factor; j++)
            bmp_put_line_fast( &bmp,
                               x_off + 70,
                               i * font_size_factor + j,
                               x_off + 120,
                               i * font_size_factor + j,
                               bmp.make_rgb(const_palette_picker(target_palette_index)[i].r,
                                            const_palette_picker(target_palette_index)[i].g,
                                            const_palette_picker(target_palette_index)[i].b
                                           )
                             );
    //------------------------------------------------------------------------
    if(show_cursors) // selected_palette_color cursor
    {
        bmp_put_line_fast( &bmp,
                           x_off + 60,
                           0,
                           x_off + 60,
                           LASERBOY_PALETTE_MAX * font_size_factor,
                           bmp.make_rgb(const_palette_picker(palette_index)[selected_palette_color].r,
                                        const_palette_picker(palette_index)[selected_palette_color].g,
                                        const_palette_picker(palette_index)[selected_palette_color].b
                                       )
                         );
    //------------------------------------------------------------------------
        for(j = 0; j < font_size_factor; j++) // selected color cursor
            bmp_put_line_fast( &bmp,
                               x_off + 40,
                               selected_palette_color * font_size_factor + j,
                               x_off + 80,
                               selected_palette_color * font_size_factor + j,
                               bmp.white
                             );
    }
    //------------------------------------------------------------------------
    if(palette_index == LASERBOY_TRUE_COLOR)
        bmp_printf(   &bmp // current frame palette name
                    , x_off
                    , y_off
                    , bmp_bg_color
                    , bmp.white
                    , 0
                    , font_size_factor
                    , "-24-bit-"
                  );
    else
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.white
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "%s"
                    , const_palette_picker(palette_index).name.c_str()
                  );

    bmp_printf(   &bmp // target palette name
                , x_off + 130
                , 0
                , bmp.white
                , bmp_bg_color
                , 1
                , font_size_factor
                , "%s"
                , const_palette_picker(target_palette_index).name.c_str()
              );

    if(show_cursors)
    {
        y_off += y_line;
        bmp_printf(   &bmp // selected color blue value
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,0,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "            %3d"
                    , const_palette_picker(palette_index)[selected_palette_color].b
                  );
        bmp_printf(   &bmp // selected color green value
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,255,0)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "        %3d"
                    , const_palette_picker(palette_index)[selected_palette_color].g
                  );
        bmp_printf(   &bmp // selected color red value
                    , x_off
                    , y_off
                    , bmp.make_rgb(255,0,0)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "    %3d"
                    , const_palette_picker(palette_index)[selected_palette_color].r
                  );
        bmp_printf(   &bmp // selected color index
                    , x_off
                    , y_off
                    , bmp.white
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "%3d"
                    , selected_palette_color
                  );
    }
}

//############################################################################
void LaserBoy_TUI::render_space()
{
    LaserBoy_color  bit_masked;

    u_int           select_start_color,
                    select_end_color  ,
                    bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    int             x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
                    y_off        =   LASERBOY_PALETTE_MAX * font_size_factor
                                   + (LASERBOY_STATS_DISPLAY_OFFSET + 24) * font_size_factor,
                    y_line       = 8 * font_size_factor;

    double          scale        = bmp.yres / 65536.0;
    //------------------------------------------------------------------------
    current_frame().render();
    //------------------------------------------------------------------------
    if(show_palettes)
        overlay_palettes();
    //------------------------------------------------------------------------
    if(show_stats)
    {
        if(current_frame().format == LASERBOY_2D_VECTOR)
        {
            if(current_frame().is_selected)
                bmp_printf(   &bmp
                            , x_off
                            , y_off
                            , bmp.make_rgb(255,255,0)
                            , bmp.make_rgb(128,0,0)
                            , 0
                            , font_size_factor
                            , "2D frame %6d"
                            , frame_index
                          );
            else
                bmp_printf(   &bmp
                            , x_off
                            , y_off
                            , bmp.make_rgb(255,255,0)
                            , bmp_bg_color
                            , 0
                            , font_size_factor
                            , "2D frame %6d"
                            , frame_index
                          );
        }
        else
        {
            //----------------------------------------------------------------
            if(current_frame().is_selected)
                bmp_printf(   &bmp
                            , x_off
                            , y_off
                            , bmp.make_rgb(255,255,0)
                            , bmp.make_rgb(128,0,0)
                            , 0
                            , font_size_factor
                            , "3D frame %6d"
                            , frame_index
                          );
            else
                bmp_printf(   &bmp
                            , x_off
                            , y_off
                            , bmp.make_rgb(255,255,0)
                            , bmp_bg_color
                            , 0
                            , font_size_factor
                            , "3D frame %6d"
                            , frame_index
                          );
        }
        //--------------------------------------------------------------------
        y_off += y_line;
        if(current_frame().is_unique)
            bmp_printf(   &bmp
                        , x_off
                        , y_off
                        , bmp.make_rgb(0,255,255)
                        , bmp_bg_color
                        , 0
                        , font_size_factor
                        , "vertices %6d"
                        , current_frame().quantity
                      );
        else
            bmp_printf(   &bmp
                        , x_off
                        , y_off
                        , bmp.make_rgb(0,255,255)
                        , bmp.make_rgb(0,128,0)
                        , 0
                        , font_size_factor
                        , "vertices %6d"
                        , current_frame().quantity
                      );
        //--------------------------------------------------------------------
        y_off += y_line;
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(255,0,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "segments %6d"
                    , current_frame().number_of_segments()
                  );
    } // end if(show_stats)
    if(show_cursors)
    {
        bit_masked = current_frame().at(current_frame().egg).bit_masked_color(signal_bit_mask);
        select_start_color =   current_frame().at(current_frame().egg).is_blank()
                             ? bmp.gray
                             :   current_frame().at(current_frame().egg).is_black()
                               ? bmp.white
                               : bmp.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b);
        //----------------------------------------------------------------
        bit_masked = current_frame().at(current_frame().spider).bit_masked_color(signal_bit_mask);
        select_end_color   =   current_frame().at(current_frame().spider).is_blank()
                             ? bmp.gray
                             :   current_frame().at(current_frame().spider).is_black()
                               ? bmp.white
                               : bmp.make_rgb(bit_masked.r, bit_masked.g, bit_masked.b);
        //----------------------------------------------------------------
        // icons
        y_off += (2 * y_line); // color vertex segment for start cursor
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(255,0,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "           %4d"
                    , current_frame().segment_index_of_vertex(current_frame().egg)
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,255,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "      %4d"
                    , current_frame().egg
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.white
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "  %3d"
                    , current_frame().at(current_frame().egg).c
                  );
        //----------------------------------------------------------------
        y_off += y_line; // RGB values for start cursor
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,0,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "            %3d"
                    , current_frame().at(current_frame().egg).b
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,255,0)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "        %3d"
                    , current_frame().at(current_frame().egg).g
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(255,0,0)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "    %3d"
                    , current_frame().at(current_frame().egg).r
                  );
        //----------------------------------------------------------------
        y_off += y_line; // color vertex segment for end cursor
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(255,0,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "           %4d"
                    , current_frame().segment_index_of_vertex(current_frame().spider)
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,255,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "      %4d"
                    , current_frame().spider
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.white
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "  %3d"
                    , current_frame().at(current_frame().spider).c
                  );
        //----------------------------------------------------------------
        y_off += y_line; // RGB values for end cursor
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,0,255)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "            %3d"
                    , current_frame().at(current_frame().spider).b
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(0,255,0)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "        %3d"
                    , current_frame().at(current_frame().spider).g
                  );
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , bmp.make_rgb(255,0,0)
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "    %3d"
                    , current_frame().at(current_frame().spider).r
                  );
        //-----------------------------------------------------------------
        y_off -= (3 * y_line);
        bmp_put_rectangle(&bmp, x_off
                              , y_off
                              , x_off + (int)(2000 * scale)
                              , y_off + (int)(2000 * scale)
                              , select_start_color
                              , 0
                         ); // start cursor icon
        //----------------------------------------------------------------
        y_off += (2 * y_line);
        bmp_put_line_fast     (&bmp, x_off + (int)(1000 * scale)
                                   , y_off
                                   , x_off + (int)(1000 * scale)
                                   , y_off + (int)(2000 * scale)
                                   , select_end_color
                              );
        bmp_put_line_fast     (&bmp, x_off + (int)(2000 * scale)
                                   , y_off + (int)(1000 * scale)
                                   , x_off
                                   , y_off + (int)(1000 * scale)
                                   , select_end_color
                              );
        bmp_put_line_fast     (&bmp, x_off
                                   , y_off
                                   , x_off + (int)(2000 * scale)
                                   , y_off + (int)(2000 * scale)
                                   , select_end_color
                              );
        bmp_put_line_fast     (&bmp, x_off + (int)(2000 * scale)
                                   , y_off
                                   , x_off
                                   , y_off + (int)(2000 * scale)
                                   , select_end_color
                              ); // end cursor icon
    //------------------------------------------------------------------------
    } // end if(show_cursors)
    //------------------------------------------------------------------------
    if(    show_TUI_clue
        && TUI_clue.size()
      )
    {
        bmp_printf(   &bmp
                    , bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor
                    , bmp.yres - 8 * font_size_factor
                    , bmp.make_rgb(255,255,0)
                    , bmp.make_rgb(63,0,63)
                    , 0
                    , font_size_factor
                    , "%s                                                     "
                    , TUI_clue.c_str()
                  );
    }
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_frame_stats()
{
    int   vertices     = current_frame().quantity,
          lit          = current_frame().number_of_lit_vertices(),
          blank        = current_frame().number_of_blank_vertices(),
          y_line       = 8 * font_size_factor,
          x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
          y_off        = bmp.yres - 19 * y_line; // 19 lines from the bottom
    u_int bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    LaserBoy_3D_short  masked;

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "Frame stats"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "frame is %s"
                , ((current_frame().is_selected) ? ("selected") : ("not selected"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "frame is %s"
                , ((current_frame().is_unique) ? ("unique") : ("not unique"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "segments ....... %d"
                , current_frame().number_of_segments()
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "vertices........ %d"
                , vertices
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "lit vertices.... %d"
                , lit
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "blank vertices.. %d"
                , blank
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "................ %f%% lit"
                , (lit/(double)vertices) * 100.0
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "................ %f%% blank"
                , (blank/(double)vertices) * 100.0
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "longest vector . %f vertices"
                , current_frame().max_distance()
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "total length ... %f vertices"
                , current_frame().total_distance()
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "sharpest angle . %f%c"
                , current_frame().max_angle() * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "total angle .... %f%c"
                , current_frame().total_angle() * one_radian
                , 0xf8
              );
    if(show_cursors)
    {
        masked = current_frame().at(current_frame().egg).bit_masked_position(signal_bit_mask);
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "egg .. %c %5d (X %6d) (Y %6d) (Z %6d)"
                    , ((current_frame().at(current_frame().egg).is_lit()) ? ('L') : ('B'))
                    , current_frame().egg
                    , masked.x
                    , masked.y
                    , masked.z
                  );
        masked = current_frame().at(current_frame().spider).bit_masked_position(signal_bit_mask);
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "spider %c %5d (X %6d) (Y %6d) (Z %6d)"
                    , ((current_frame().at(current_frame().spider).is_lit()) ? ('L') : ('B'))
                    , current_frame().spider
                    , masked.x
                    , masked.y
                    , masked.z
                  );
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "select center  (X %6.0f) (Y %6.0f) (Z %6.0f)"
                    , current_frame().selected_segment().find_center().x
                    , current_frame().selected_segment().find_center().y
                    , current_frame().selected_segment().find_center().z
                  );
    }
    if(show_fulcrum)
    {
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "fulcrum ...... (X %6.0f) (Y %6.0f) (Z %6.0f)"
                    , fulcrum.x
                    , fulcrum.y
                    , fulcrum.z
                  );
    }
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_frame_set_stats()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = bmp.yres - 7 * y_line;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    count_frames();
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "Frame set stats"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , next_hue()
                , bmp_bg_color
                , 0
                , font_size_factor
                , "frames    %d"
                , number_of_frames()
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , next_hue()
                , bmp_bg_color
                , 0
                , font_size_factor
                , "2D frames %d"
                , num_2D_frames
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , next_hue()
                , bmp_bg_color
                , 0
                , font_size_factor
                , "3D frames %d"
                , num_3D_frames
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , next_hue()
                , bmp_bg_color
                , 0
                , font_size_factor
                , "verticies %d"
                , total_vertices()
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_palette_set_stats()
{
    int    i,
           y_line       = 8 * font_size_factor,
           x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp
                , x_off
                , 0
                , bmp.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "Palette set stats "
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , next_hue()
                , bmp_bg_color
                , 0
                , font_size_factor
                , "total palettes %d "
                , number_of_palettes()
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , next_hue()
                , bmp_bg_color
                , 0
                , font_size_factor
                , "added palettes %d"
                , number_of_palettes() - LASERBOY_BUILT_IN_PALETTES
              );
    for(i = 0; i < number_of_palettes(); i++)
    {
        y_off += y_line;
        bmp_printf(   &bmp
                    , x_off
                    , y_off
                    , next_hue()
                    , bmp_bg_color
                    , 0
                    , font_size_factor
                    , "%s          "
                    , LaserBoy_palette_set::at(i).name.c_str()
                  );
    }
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_program_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "Welcome to " LASERBOY_VERSION " !!!"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      i input  file"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      o output file"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      t signal bit resolution"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      u user interface visuals"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      v view keys & apply"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    ` ~ auto show"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    - _ auto orbit"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    = + auto show & orbit"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    %c %c frame"
                , 0x1b
                , 0x1a
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    %c %c target palette"
                , 0x18
                , 0x19
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      9 first frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      0 last frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      ( set | unset frames-( select"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      ) set | unset )-frames select"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Space] select | unselect frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Bkspc] delete current frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      p palette   transforms"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      h frame     transforms"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      j frame set transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      k draw, color, blank  (vertices)"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      l move, scale, rotate (segments)"
              );
    if(bg_file_name != ".")
    {
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "      ; show | hide background bitmap"
                  );
    }
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      b wave utilities"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    / ? frame set stats | frame stats"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [Tab] system settings"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [Esc] quit LaserBoy, saving OUT.ild"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Enter] show | hide help"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_please_wait()
{
    int x_off  = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off,
        y_line = 8 * font_size_factor;

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp.make_rgb(0,0,192), 0, font_size_factor
                , "please wait ........................................"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_in_file_type_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "file input type"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 ild"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 dxf"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 bitmap"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_in_ild_action_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "ild file input action"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 replace frame set"
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 prepend to frame set"
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 insert after current frame"
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 append to frame set"
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 superimpose frame(s) into frame set"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_in_dxf_action_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "DXF file input action"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 replace frame set with dxf file"
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 append  frame set with dxf file"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "frame set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 replace frame set with dxf directory"
              );

    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 append  frame set with dxf directory"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_in_wave_action_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "wave file input action"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy formatted wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 replace current frame set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 append to current frame set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 open as unformatted wave"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "unformatted wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 replace current frame set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 append to current frame set"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 import ADAT wave set"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_in_bmp_action_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "bitmap file input action"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 open as frame color map"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 open as frame set color map"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 open as frame mask"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 open as frame set mask"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 open into frame background"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 clear frame background"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_if_not_z_options()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "signal option if not z"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 nothing"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 monochrome TTL (on or off)"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 monochrome simple average of intensity"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 monochrome bitwise OR of intensity"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 monochrome weighted rgb gray scale"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_out_file_type_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "Output file type"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 ild save current frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 ild save selected frames"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 ild save all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 ild generate & save frame effect"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 ild generate & save frame set effect"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 dxf save current frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    7 dxf save directory selected frames"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    8 dxf save directory all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    9 wave animation %3.3f fps of all frames"
                , frames_per_second
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    0 wave %6.2f second still of current frame"
                , still_frame_duration
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    a C++ save current frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    b C++ save selected frames"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    c C++ save all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    d bmp save %d%c pxl of current frame"
                , output_bmp_size
                , 0xfd
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    e bmp save directory %d%c pxl selected frames"
                , output_bmp_size
                , 0xfd
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    f bmp save directory %d%c pxl all frames"
                , output_bmp_size
                , 0xfd
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    g ASCII palette save current frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    h ASCII palette save target palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    i save system settings [LaserBoy.wtf]"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] cancel action"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_settings_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 insignificant angle in degrees ....... %6.2f%c"
                , insignificant_angle * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 insignificant distance in vertices ... %6.2f"
                , insignificant_distance
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 max optimized lit vector in vertices . %6.2f"
                , lit_delta_max
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 max optimized blank vector in vertices %6.2f"
                , blank_delta_max
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 maximum dwell in microseconds ........ %6.2f"
                , max_dwell_microsec
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 lit dwell overhang in samples ........ %d"
                , lit_dwell_overhang
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    7 dwell on dot in samples .............. %d"
                , dwell_on_dot
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    8 move vertices per key hit ............ %d"
                , displacement_step
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    9 rotate degrees per key hit ........... %3.2f%c"
                , rotation_step * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    0 scale percent per key hit ............ %3.2f%%"
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    a rendered arc step in degrees ......... %3.2f%c"
                , rendered_arc_angle
                , 0xf8
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    b frames per second, 0 for once each ... %3.2lf"
                , frames_per_second
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    c still frame durration in seconds ..... %6.2f"
                , still_frame_duration
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    d output wave sample rate .............. %d"
                , sample_rate
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    e unframed wave saples per frame ....... %d"
                , unframed_wave_stride
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    f wave signal if not z ................. %s"
                , LaserBoy_wave_signal_id_to_name(signal_if_not_z).c_str()
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    g output bitmap size in square pixels .. %d"
                , output_bmp_size
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    h display font size factor ............. %d"
                , font_size_factor
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    i display background color ......... red %d"
                , bg_color.r
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    j display background color ....... green %d"
                , bg_color.g
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    k display background color ........ blue %d"
                , bg_color.b
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    l display background gray scale ........ %d"
                , (bg_color.r + bg_color.g + bg_color.b) / 3
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    m recolor palette span factor .......... %6.2f"
                , recolor_span_factor
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    n black level for palette conversion  .. %d"
                , black_threshold
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] exit settings menu"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_veiw_menu()
{
    int    x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off,
           y_line       =  8 * font_size_factor,
           x_tab        = 26 * y_line,
           x_tab2       =  9 * y_line;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "press & hold view keys, [Shift] reverse"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "1 ! orbit X   %3.2f%c"
                , rotation_step * one_radian
                , 0xf8
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "5 %% front | back"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "2 @ orbit  Y  %3.2f%c"
                , rotation_step * one_radian
                , 0xf8
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "6 ^  left | right"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "3 # orbit   Z %3.2f%c"
                , rotation_step * one_radian
                , 0xf8
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "7 &   top | bottom"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "4 $ orbit XYZ %3.2f%c"
                , rotation_step * one_radian
                , 0xf8
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "8 * orbit & frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "q Q pan X  %5d vertices"
                , displacement_step
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "a A zoom X  %3.2f%%"
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "w W pan  Y %5d vertices"
                , displacement_step
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "s S zoom  Y %3.2f%%"
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "e E pan XY %5d vertices"
                , displacement_step
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "d D zoom XY %3.2f%%"
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "r R center"
              );
    bmp_printf(   &bmp
                , x_tab
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "f F 1:1"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "current view"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_tab2
                , y_off
                , bmp.make_rgb(255,0,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "X"
              );
    bmp_printf(   &bmp
                , x_tab2 * 2
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "Y"
              );
    bmp_printf(   &bmp
                , x_tab2 * 3
                , y_off
                , bmp.make_rgb(0,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "Z"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , " angle                               degrees"
              );
    bmp_printf(   &bmp
                , x_tab2
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.x * one_radian
              );
    bmp_printf(   &bmp
                , x_tab2 * 2
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.y * one_radian
              );
    bmp_printf(   &bmp
                , x_tab2 * 3
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp.make_rgb(0,0,80)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.z * one_radian
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "offset                               vertices"
              );
    bmp_printf(   &bmp
                , x_tab2
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-8.0f"
                , view_offset.x
              );
    bmp_printf(   &bmp
                , x_tab2 * 2
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-8.0f"
                , view_offset.y
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , " scale"
              );
    bmp_printf(   &bmp
                , x_tab2
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_scale.x
              );
    bmp_printf(   &bmp
                , x_tab2 * 2
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_scale.y
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "      o O apply angle  to frame | all"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(255,0,255)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "      t T apply offset to frame | all"
              );
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.make_rgb(0,255,0)
                , bmp_bg_color
                , 0
                , font_size_factor
                , "      g G apply scale  to frame | all"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp
                , x_off
                , y_off
                , bmp.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "  [Esc] exit view settings"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_visuals_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy toggle UI visuals settings"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_vertices)?("--> 1 vertices"):("    1 vertices"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_blanking)?("--> 2 blanking"):("    2 blanking"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_palettes)?("--> 3 palettes"):("    3 palettes"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_stats)?("--> 4 stats"):("    4 stats"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_fixed_bounds)?("--> 5 fixed bounds"):("    5 fixed bounds"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_fixed_origin)?("--> 6 fixed origin"):("    6 fixed origin"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_floating_axis)?("--> 7 floating axis"):("    7 floating axis"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_floating_bounds)?("--> 8 floating bounds"):("    8 floating bounds"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_TUI_clue)?("--> 9 TUI clue"):("    9 TUI clue"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    0 show | hide all visuals"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "toggle LaserBoy system settings"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((show_effects_generation)?("--> a show effect rendering"):("    a show effect rendering"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((auto_flatten_z)?("--> b flatten Z for wave output"):("    b flatten Z for wave output"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((no_equivalent_vectors)?("--> c remove equal vectors"):("    c remove equal vectors"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((no_dots)?("--> d remove dots (not lines)"):("    d remove dots (not lines)"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((invert_wave_output)?("--> e inverted wave output"):("    e inverted wave output"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((auto_optimize)?("--> f optimize before save"):("    f optimize before save"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((loop_animation)?("--> g loop animation"):("    g loop animation"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((approximate_frame_rate)?("--> h animate at approximated frame rate"):("    h animate at approximated frame rate"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((auto_scale_dxf)?("--> i auto scale dxf import"):("    i auto scale dxf import"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((maintain_dxf_origin)?("--> j maintain dxf origin"):("    j maintain dxf origin"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((black_dwell_vertices)?("--> k black dwell"):("    k black dwell"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((no_black_match)?("--> l no black palette match"):("    l no black palette match"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((allow_lit_black)?("--> m allow lit black"):("    m allow lit black"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                   , ((dumb_file_format)?("--> n save ild format 4,5"):("    n save ild format 4,5"))
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] exit LaserBoy UI visuals settings"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_palette_tranz_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "palette transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      1 span X with palette"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      2 span Y with palette"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      3 span Z with palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      4 span  vectors with palette"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      5 index vectors with palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      6 span  lit segments with palette"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      7 index lit segments with palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      8 random color vectors with palette"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      9 random color lit segments with palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      0 set to target palette by index"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      a best match to target palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      b strip color information (white)"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      c strip color information (r|g|b)"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      d strip color information gray scale"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      e convert true color to palette (if possible)"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      f bit reduce true color to 332 palette"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      g best reduce true color to palette"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    h H best reduce all frames to 1 palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      i promote to true color"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    j J clear unused palettes"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Shift] select to apply to all frames"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "  [Esc] exit palette transforms"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_frame_set_tranz_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "frame set transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      1 delete selected"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      2 trim to selected"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      3 delete repeats from wave import"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      4 copy selected to begining"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      5 copy selected to end"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      6 move selected to begining"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      7 move selected to end"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      8 add empty frame to begining"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      9 add empty frame after current"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      0 add empty frame to end"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      a reverse frame set order"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      b reverse order of selected frames"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      c make current frame first"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      d randomize frame set order"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      e delete every n-th frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      > explode frame into segment set"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      < collapse selected frames into one"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      ! explode all frames into segments"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Space] select | unselect frame"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Enter] show | hide help"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "  [Esc] exit frame set transforms"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_frame_tranz_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "frame transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      1 flatten Z (force to 2D)"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      2 Z order vertices"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      3 reverse vector order"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      4 mirror"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      5 flip"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      6 quarter turn CW"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      7 quarter turn CCW"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      8 accentuate the positive"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      9 eliminate the negative"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      0 impose bit resolution"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      a convert black to blank"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      b convert to dots"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      c remove blank span"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      d remove angle dwell"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      e remove dots (not lines)"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      f reduce vectors"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      g blank equivalent vectors"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      h bond segments (reorder)"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      i minimize"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      j fracture segments"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      k reorder segments"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      l add angle dwell"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      m add lit distance span"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      n add blank distance span"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      o enhance dots"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      p optimize"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Shift] select applies to all frames"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Enter] show | hide help"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "  [Esc] exit frame transforms"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_draw_color_blank_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "draw, color, blank mode entered"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    \\ | %c to 0 | %c to end"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    [ { +- %c   one vertex"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    ] } +-   %c one vertex"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    i I +- %c   lit segment"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    o O +-   %c lit segment"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    u U +- %c-%c lit segment"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    a A unblank all vectors | frames"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    b B blank | unblank %c-%c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    n N bond black | unbond unblack %c-%c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    p P +- palette color index"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      c color %c-%c with palette index"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      C get palette index from %c color"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      h replace black in %c-%c with palette index"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    X Y Z  move   %c + %d"
                , 0x0f
                , displacement_step
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    X Y Z  move   %c - %d"
                , 0x0f
                , displacement_step
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    q w e  set    %c coordinates X Y Z"
                , 0x0f
                , displacement_step
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    f F    rotate %c +- %3.2f%c on Z axis"
                , 0x0f
                , rotation_step * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    g G    rotate %c +- %3.2f%c on Y axis"
                , 0x0f
                , rotation_step * one_radian
                , 0xf8
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    m M +- %3.2f%% magnitude of %c"
                , magnitude_step * 100.00
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      l draw line    %c to %c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      k draw polygon %c to %c radius"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      d %3.2f%c circular  arc, %c to %c radius"
                , rendered_arc_angle
                , 0xf8
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      D %3.2f%c eliptical arc, %c to %c radius"
                , rendered_arc_angle
                , 0xf8
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      t render text on line  %c to %c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    , < reorder vertices from %c | %c"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      ' convert %c-%c to dots"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      . split vector in line or add vector to end"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      s add blank vertex at %c (cut segment)"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Bkspc] delete vertex at %c"
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    / ? show frame | set stats"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "%c %c %c %c and view orbit keys are active"
                , 0x1b
                , 0x1a
                , 0x18
                , 0x19
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Enter] show | hide help"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "  [Esc] exit draw, color blank mode"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_move_scale_rotate_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "move, scale, rotate mode entered"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    \\ | %c to 0 | %c to end"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    [ { +- %c   one vertex"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    ] } +-   %c one vertex"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    i I +- %c   lit segment"
                , 0x01
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    o O +-   %c lit segment"
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    u U +- %c-%c lit segment"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  X Y Z  move  %c-%c + %d"
                , 0x01
                , 0x0f
                , displacement_step
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  X Y Z  move  %c-%c - %d"
                , 0x01
                , 0x0f
                , displacement_step
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  q w e  set center coordinates X Y Z"
                , 0x0f
                , displacement_step
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    f F rotate %c-%c +- %3.2f%c Z axis on center"
                , 0x01
                , 0x0f
                , rotation_step * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    g G rotate %c-%c +- %3.2f%c Y axis on center"
                , 0x01
                , 0x0f
                , rotation_step * one_radian
                , 0xf8
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    9 ( scale  %c-%c X   +- %3.2f%% on center"
                , 0x01
                , 0x0f
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    0 ) scale  %c-%c  Y  +- %3.2f%% on center"
                , 0x01
                , 0x0f
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    - _ scale  %c-%c   Z +- %3.2f%% on center"
                , 0x01
                , 0x0f
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    = + scale  %c-%c XYZ +- %3.2f%% on center"
                , 0x01
                , 0x0f
                , magnitude_step * 100.00
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  a s d  move  fulcrum X Y Z + %d"
                , displacement_step
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  A S D  move  fulcrum X Y Z - %d"
                , displacement_step
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  j k l  set   fulcrum coordinates X Y Z"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    , .  put   fulcrum on egg | spider"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    r R rotate  %c-%c +- %3.2f%c Z axis on fulcrum"
                , 0x01
                , 0x0f
                , rotation_step * one_radian
                , 0xf8
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    t T rotate  %c-%c +- %3.2f%c Y axis on fulcrum"
                , 0x01
                , 0x0f
                , rotation_step * one_radian
                , 0xf8
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    v V scale   %c-%c X   +- %3.2f%% from fulcrum"
                , 0x01
                , 0x0f
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    b B scale   %c-%c  Y  +- %3.2f%% from fulcrum"
                , 0x01
                , 0x0f
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    n N scale   %c-%c   Z +- %3.2f%% from fulcrum"
                , 0x01
                , 0x0f
                , magnitude_step * 100.00
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    m M scale   %c-%c XYZ +- %3.2f%% from fulcrum"
                , 0x01
                , 0x0f
                , magnitude_step * 100.00
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      < reverse %c-%c vector order"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "      c copy    %c-%c (in place) to end"
                , 0x01
                , 0x0f
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    / ? frame | set stats"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "%c %c %c %c and view orbit keys are active"
                , 0x1b
                , 0x1a
                , 0x18
                , 0x19
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "[Enter] show | hide help"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "  [ESC] exit move, scale, rotate mode"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_wave_utilities_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "wave utilities"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 wave display settings (monitor)"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 show LaserBoy wave header information"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 show wave on screen"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    4 set desired wave sample offsets"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    5 apply offsets to wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    6 overwrite offsets in LaserBoy wave header"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    7 invert global polarity of wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    8 invert signal polarities of LaserBoy wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    9 clear signal polarity list in LaserBoy wave"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    0 split wave to XY, r, g, b, i"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    a join XY, r, g, b, i waves to multi channel wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    b join XY, r+g, b, waves to multi channel wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    c join XY, i, L, R waves to multi channel wave"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    d add stereo audio to wave"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] exit wave utilities"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_show_wave_options_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "show wave options"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    1 show LaserBoy wave"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    2 show LaserBoy wave as unformatted"
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "    3 show unformatted wave"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[ESC] exit show wave options"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_wave_show_settings_menu()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "wave show settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_show_vertices)?("--> 1 vertices"):("    1 vertices"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_show_blanks)?("--> 2 blanking"):("    2 blanking"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_real_time)?("--> 3 real time"):("    3 real time"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_show_inverted)?("--> 4 inverted"):("    4 inverted"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_clean_screen)?("--> 5 clean screen"):("    5 clean screen"))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((wave_loop)?("--> 6 loop"):("    6 loop"))
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] exit wave show settings"
              );
    return;
}

//############################################################################
bool LaserBoy_TUI::overlay_wave_stats(string file_name)
{
    int            i,
                   y_line       = 8 * font_size_factor,
                   x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
                   y_off        = x_off;
    u_int          bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    ifstream       in;
    //------------------------------------------------------------------------
    in.open(file_name.c_str(), ios::in | ios::binary);
    LaserBoy_wave_header header(in);
    in.close();
    if(header.version == "!LaserBoy!")
    {
        bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                    , "[any key] %s failed to open"
                    , file_name.c_str()
                  );
        error = LASERBOY_FILE_OPEN_FAILED;
        return false;
    }
    //------------------------------------------------------------------------
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy wave stats                                 "
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , file_name.c_str()
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "LaserBoy wave header version : %s"
                , header.version.c_str()
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "num_samples ..........: %d"
                , header.num_samples
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "sample_rate ..........: %d"
                , header.sample_rate
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "number of channels ...: %d"
                , header.num_channels
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "bits per sample ......: %d"
                , header.bits_per_sample
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "playing time .........: %f seconds"
                , ((double)header.num_samples / header.sample_rate)
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((header.LaserBoy_wave_mode & LASERBOY_WAVE_POSITIVE) ? ("wave polarity ........: positive signal") : ("wave polarity ........: negative signal"))
              );
    if(header.LaserBoy_wave_mode & LASERBOY_WAVE_OPTIMIZED)
    {
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "lit dwell overhang ...: %d"
                    , header.parms.lit_dwell_overhang
                  );
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "lit distance delta max: %6.2f"
                    , header.parms.lit_delta_max
                  );
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "blank dist delta max .: %6.2f"
                    , header.parms.blank_delta_max
                  );
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "longest dwell u_sec ..: %6.2f"
                    , header.parms.max_dwell_microsec
                  );
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "insignificant distance: %6.2f"
                    , header.parms.insignificant_distance
                  );
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "insignificant angle ..: %6.2f"
                    , header.parms.insignificant_angle
                  );
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "frames per second ....: %6.2f"
                    , header.parms.frames_per_second
                  );
    }
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "ch: sig : contents     : LSB subcode  : offset : res"
              );
    for(i = 0; i < header.num_channels; i++)
    {
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "%2d: %s : %s : %s :   %4d : %3d"
                    , i
                    , ((header.signal_id[i] >= 0) ? ("   ") : ("neg"))
                    , LaserBoy_wave_signal_id_to_name(header.signal_id[i]).c_str()
                    , LaserBoy_wave_LSB_tag_to_name(header.LSB_tag[i]).c_str()
                    , header.offset[i]
                    , header.resolution[i]
                  );
    }
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "press any key to continue"
                , ((double)header.num_samples / header.sample_rate)
              );
    return true;
}

//############################################################################
short LaserBoy_TUI::overlay_wave_signal_polarities(string file_name, int polarity[8])
{
    int            i,
                   y_line       = 8 * font_size_factor,
                   x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
                   y_off        = x_off;
    u_int          bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    ifstream       in;
    //------------------------------------------------------------------------
    in.open(file_name.c_str(), ios::in | ios::binary);
    LaserBoy_wave_header header(in);
    in.close();
    if(header.version == "!LaserBoy!")
    {
        bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                    , "[any key] %s failed to open"
                    , file_name.c_str()
                  );
        error = LASERBOY_FILE_OPEN_FAILED;
        return 0;
    }
    //------------------------------------------------------------------------
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy wave stats                                 "
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , file_name.c_str()
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , ((header.LaserBoy_wave_mode & LASERBOY_WAVE_POSITIVE) ? ("wave polarity ........: positive signal") : ("wave polarity ........: negative signal"))
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "ch: sig : contents"
              );
    for(i = 0; i < header.num_channels; i++)
    {
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "%2d: %s : %s"
                    , i
                    , ((header.signal_id[i] >= 0) ? ("   ") : ("neg"))
                    , LaserBoy_wave_signal_id_to_name(header.signal_id[i]).c_str()
                  );
    }
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy wave signal polarity matrix"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [0] X          %s"
                , ((polarity[0] == -1) ? ("") : ((polarity[0]) ? ("invert") : ("unchanged")))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [1] Y          %s"
                , ((polarity[1] == -1) ? ("") : ((polarity[1]) ? ("invert") : ("unchanged")))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [2] red        %s"
                , ((polarity[2] == -1) ? ("") : ((polarity[2]) ? ("invert") : ("unchanged")))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [3] green      %s"
                , ((polarity[3] == -1) ? ("") : ((polarity[3]) ? ("invert") : ("unchanged")))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [4] blue       %s"
                , ((polarity[4] == -1) ? ("") : ((polarity[4]) ? ("invert") : ("unchanged")))
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [5] Z or other %s"
                , ((polarity[5] == -1) ? ("") : ((polarity[5]) ? ("invert") : ("unchanged")))
              );
    if(header.num_channels == 8)
    {
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "  [6] L audio    %s"
                    , ((polarity[6] == -1) ? ("") : ((polarity[6]) ? ("invert") : ("unchanged")))
                  );
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "  [7] R audio    %s"
                    , ((polarity[7] == -1) ? ("") : ((polarity[7]) ? ("invert") : ("unchanged")))
                  );
    }
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel"
              );
    return header.num_channels;
}

//############################################################################
void LaserBoy_TUI::overlay_wave_offsets()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy wave sample offset settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [0] X          %d"
                , wave_offsets[0]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [1] Y          %d"
                , wave_offsets[1]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [2] red        %d"
                , wave_offsets[2]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [3] green      %d"
                , wave_offsets[3]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [4] blue       %d"
                , wave_offsets[4]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [5] Z or other %d"
                , wave_offsets[5]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [6] L audio    %d"
                , wave_offsets[6]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [7] R audio    %d"
                , wave_offsets[7]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_bit_resolution_settings()
{
    int    y_line       = 8 * font_size_factor,
           x_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = x_off;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "LaserBoy signal bit resolution settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [0] X       %d"
                , 16 - signal_bit_mask[0]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [1] Y       %d"
                , 16 - signal_bit_mask[1]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [5] Z       %d"
                , 16 - signal_bit_mask[5]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [2] red     %d"
                , 16 - signal_bit_mask[2]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [3] green   %d"
                , 16 - signal_bit_mask[3]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [4] blue    %d"
                , 16 - signal_bit_mask[4]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [6] L audio %d"
                , 16 - signal_bit_mask[6]
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                , "  [7] R audio %d"
                , 16 - signal_bit_mask[7]
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "[Esc] cancel"
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_file_list(string directory)
{
    bool            found_one    = false;
    int             i,
                    y_line       = 8 * font_size_factor,
                    x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
                    y_off        = y_line;
    u_int           bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    DIR            *pDIR;
    struct dirent  *pDirEnt;

    pDIR = opendir(directory.c_str());
    if(pDIR == NULL)
        return;
    bmp_printf(   &bmp
                , x_off
                , 0
                , bmp.white
                , bmp_bg_color
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "Palette set stats "
              );
    pDirEnt = readdir(pDIR);
    while(pDirEnt != NULL)
    {
        if(    strcmp(pDirEnt->d_name, ".")  // not the dot
            && strcmp(pDirEnt->d_name, "..") // not the dot dot
          )
        {
            bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor, "%s                  "
                        , pDirEnt->d_name
                      );
            y_off += y_line;
            found_one = true;
        }
        pDirEnt = readdir(pDIR);
    }
    closedir(pDIR);
    if(!found_one)
    {
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "none found        "
                  );
        y_off += y_line;
    }
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    y_off += y_line;
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_directory_list(string directory)
{
    bool            found_one    = false;
    int             i,
                    y_line       = 8 * font_size_factor,
                    x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
                    y_off        = y_line;
    u_int           bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);
    DIR            *pDIR;
    struct dirent  *pDirEnt;

    pDIR = opendir(directory.c_str());
    if(pDIR == NULL)
        return;
    bmp_printf(   &bmp, x_off, 0, bmp.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "Palette set stats "
              );
    pDirEnt = readdir(pDIR);
    while(pDirEnt != NULL)
    {
        if(!strstr(pDirEnt->d_name, ".")) // if there is no dot in the name
        {
            bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                        , "%s                  "
                        , pDirEnt->d_name
                      );
            y_off += y_line;
            found_one = true;
        }
        pDirEnt = readdir(pDIR);
    }
    closedir(pDIR);
    if(!found_one)
    {
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "none found        "
                  );
        y_off += y_line;
    }
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_frame_effects()
{
    int    i,
           y_line       = 8 * font_size_factor,
           x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, 0, bmp.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "frame effects"
              );
    if(frame_effect_names.size() > 0)
    {
        for(i = 0; i < (int)frame_effect_names.size(); i++)
        {
            y_off += y_line;
            bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor, "%s          "
                        , frame_effect_names[i].c_str()
                      );
        }
    }
    else
    {
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "no frame effects loaded"
                  );
    }
    return;
}

//############################################################################
void LaserBoy_TUI::overlay_frame_set_effects()
{
    int    i,
           y_line       = 8 * font_size_factor,
           x_off        = bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off        = LASERBOY_STATS_DISPLAY_OFFSET * font_size_factor;
    u_int  bmp_bg_color = bmp.make_rgb(bg_color.r, bg_color.g, bg_color.b);

    bmp_printf(   &bmp, x_off, 0, bmp.white, bmp_bg_color, 0, font_size_factor
                , "                  "
              );
    bmp_printf(   &bmp, x_off, y_off, bmp.white, bmp_bg_color, 0, font_size_factor
                , "frame set effects"
              );
    if(frame_set_effect_names.size() > 0)
    {
        for(i = 0; i < (int)frame_set_effect_names.size(); i++)
        {
            y_off += y_line;
            bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor, "%s          "
                        , frame_set_effect_names[i].c_str()
                      );
        }
    }
    else
    {
        y_off += y_line;
        bmp_printf(   &bmp, x_off, y_off, next_hue(), bmp_bg_color, 0, font_size_factor
                    , "no frame set effects loaded"
                  );
    }
    return;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
