//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "LaserBoy_GUI_base.hpp"

//############################################################################
LaserBoy_frame_set::LaserBoy_frame_set(LaserBoy_space* ps,
                                       LaserBoy_vertex(*F)(int, int),
                                       int vertices_per_frame,
                                       int num_frames
                                      )
 : p_space            (ps)
 , frame_select_start (-1)
 , frame_select_end   (-1)
 , error              (LASERBOY_OK)
 , null_frame         (ps)
 , intro              (ps)
{
    null_frame = NULL_frame(ps);
    reserve(num_frames);
    for(frame_index = 0; frame_index < num_frames; frame_index++)
        push_back(LaserBoy_frame(p_space, F, vertices_per_frame, frame_index));
    renumber();
    frame_index = 0;
    error = LASERBOY_OK;
}

//############################################################################
LaserBoy_frame_set& LaserBoy_frame_set::operator += (const LaserBoy_frame& frame)
{
    if(p_space->show_effects_generation)
    {
        frame.render();
        p_space->GUI->display_space();
    }
    push_back(frame);
    renumber();
    return *this;
}

//############################################################################
bool LaserBoy_frame_set::from_ild_file(string file, bool append)
{
    ifstream in(file.c_str(), ios::in | ios::binary);

    if(in.is_open())
    {
        frame_index = 0;
        error       = LASERBOY_OK;
        error       = from_ifstream_ild(in, append);
        in.close();
    }
    else
        error = LASERBOY_FILE_OPEN_FAILED;

    if(error)
        return false;
    return true;
}

//############################################################################
bool LaserBoy_frame_set::superimpose_from_ild_file(string file)
{
    int                 i;
    LaserBoy_frame_set  overlay(p_space),
                        sum    (p_space);

    if(    overlay.from_ild_file(file)
        && overlay.number_of_frames()
      )
    {
        if(overlay.number_of_frames() > 1 && number_of_frames() > 1)
        {
            if(number_of_frames() > overlay.number_of_frames())
            {
                for(i = 0; i < overlay.number_of_frames(); i++)
                    sum += (frame(i) + overlay.frame(i));

                for(i = overlay.number_of_frames(); i < number_of_frames(); i++)
                    sum += frame(i);
            }
            else if(number_of_frames() < overlay.number_of_frames())
            {
                for(i = 0; i < number_of_frames(); i++)
                    sum += (frame(i) + overlay.frame(i));

                for(i = number_of_frames(); i < overlay.number_of_frames(); i++)
                    sum += overlay.frame(i);
            }
            else
            {
                for(i = 0; i < number_of_frames(); i++)
                    sum += (frame(i) + overlay.frame(i));
            }
        }
        else if(overlay.number_of_frames() == 1)
        {
            for(i = 0; i < number_of_frames(); i++)
                sum += (frame(i) + overlay.frame(0));
        }
        else if(number_of_frames() == 1)
        {
            for(i = 0; i < overlay.number_of_frames(); i++)
                sum += (frame(0) + overlay.frame(i));
        }
        *this = sum;
        return true;
    }
    return false;
}


//############################################################################
bool LaserBoy_frame_set::from_dxf_file(string file, bool append)
{
    ifstream in(file.c_str(), ios::in | ios::binary);

    if(in.is_open())
    {
        frame_index = 0;
        error       = LASERBOY_OK;
        error       = from_ifstream_dxf(in, append);
        in.close();
    }
    else
        error = LASERBOY_FILE_OPEN_FAILED;

    if(error)
        return false;

    if(at(0).find_rgb_in_palette(p_space->palette_picker(LASERBOY_DXF)))
        at(0).palette_index = LASERBOY_DXF;
    return true;
}

//############################################################################
bool LaserBoy_frame_set::from_dxf_directory(string dir, bool append)
{
    DIR                    *pDIR;
    struct dirent          *pDirEnt;
    vector<string>          file_list;
    //------------------------------------------------------------------------
    pDIR = opendir(dir.c_str());
    //------------------------------------------------------------------------
    if(pDIR == NULL)
        error = LASERBOY_FILE_OPEN_FAILED;
    else
    {
        pDirEnt = readdir(pDIR);
        while(pDirEnt != NULL)
        {
            if(    strcmp(pDirEnt->d_name, "." ) // not the dot
                && strcmp(pDirEnt->d_name, "..") // not the dot dot
              )
                file_list.push_back(pDirEnt->d_name);
            pDirEnt = readdir(pDIR);
        }
        closedir(pDIR);
        if(file_list.size())
        {
            bool  valid_data = false;
            int   i, j;

            LaserBoy_real_segment      real_vertices(p_space);
            LaserBoy_real_segment_set  real_segments(p_space);
            LaserBoy_vertex            vertex;
            LaserBoy_frame             frame(p_space);
            //----------------------------------------------------------------
            frame.format        = LASERBOY_3D_VECTOR ;
            frame.palette_index = LASERBOY_TRUE_COLOR;
            //----------------------------------------------------------------
            if(!append)
            {
                clear();
                num_2D_frames = 0;
                num_3D_frames = 0;
            }
            //----------------------------------------------------------------
            for(i = 0; i < (int)file_list.size(); i++)
            {
                ifstream in((dir + '/' + file_list[i]).c_str(), ios::in);
                if(in.is_open())
                {
                    real_vertices.from_ifstream_dxf(in);
                    if(real_vertices.size() > 2) // there is a picture in there
                        valid_data = true;
                    real_segments.push_back(real_vertices);
                    real_vertices.clear();
                    in.close();
                }
            }
            //----------------------------------------------------------------
            if(valid_data)
            {
                double              real_size   =  0.0,
                                    real_scale  =  1.0;
                LaserBoy_3D_double  real_min    =  DBL_MAX,
                                    real_max    = -DBL_MAX,
                                    real_offset =  0.0;
                if(p_space->auto_scale_dxf)
                {
                    for(j = 0; j < (int)real_segments.size(); j++)
                    {
                        if(real_segments[j].size() > 2)
                        {
                            for(i = 2; i < (int)real_segments[j].size(); i++)
                            {
                                if(real_segments[j][i].x > real_max.x)    real_max.x = real_segments[j][i].x;
                                if(real_segments[j][i].x < real_min.x)    real_min.x = real_segments[j][i].x;
                                if(real_segments[j][i].y > real_max.y)    real_max.y = real_segments[j][i].y;
                                if(real_segments[j][i].y < real_min.y)    real_min.y = real_segments[j][i].y;
                                if(real_segments[j][i].z > real_max.z)    real_max.z = real_segments[j][i].z;
                                if(real_segments[j][i].z < real_min.z)    real_min.z = real_segments[j][i].z;
                            }
                        }
                    }
                    if(p_space->maintain_dxf_origin)
                    {
                        if(fabs(real_max.x) > real_size)    real_size = fabs(real_max.x);
                        if(fabs(real_min.x) > real_size)    real_size = fabs(real_min.x);
                        if(fabs(real_max.y) > real_size)    real_size = fabs(real_max.y);
                        if(fabs(real_min.y) > real_size)    real_size = fabs(real_min.y);
                        if(fabs(real_max.z) > real_size)    real_size = fabs(real_max.z);
                        if(fabs(real_min.z) > real_size)    real_size = fabs(real_min.z);
                        real_size *= 2;
                    }
                    else // find the new center of the universe
                    {
                        real_offset.x = (real_max.x - ((real_max.x - real_min.x) / 2));
                        real_offset.y = (real_max.y - ((real_max.y - real_min.y) / 2));
                        real_offset.z = (real_max.z - ((real_max.z - real_min.z) / 2));
                        if(fabs(real_max.x - real_min.x) > real_size)    real_size = fabs(real_max.x - real_min.x);
                        if(fabs(real_max.y - real_min.y) > real_size)    real_size = fabs(real_max.y - real_min.y);
                        if(fabs(real_max.z - real_min.z) > real_size)    real_size = fabs(real_max.z - real_min.z);
                    }
                    if(real_size)
                        real_scale = 65000.0 / real_size;
                    else
                        real_scale = 1.0;
                } // end if(auto_scale_dxf)
                //------------------------------------------------------------
                reserve(real_segments.size());
                //------------------------------------------------------------
                for(j = 0; j < (int)real_segments.size(); j++)
                {
                    if(real_segments[j].size() > 2)
                    {
                        for(i = 2; i < (int)real_segments[j].size(); i++)
                        {
                            vertex.x = (short)((real_segments[j][i].x - real_offset.x) * real_scale);
                            vertex.y = (short)((real_segments[j][i].y - real_offset.y) * real_scale);
                            vertex.z = (short)((real_segments[j][i].z - real_offset.z) * real_scale);
                            vertex.r = real_segments[j][i].r;
                            vertex.g = real_segments[j][i].g;
                            vertex.b = real_segments[j][i].b;
                            if(real_segments[j][i].is_blank())
                                vertex.blank();
                            else
                                vertex.unblank();
                            vertex.set_unique();
                            frame.push_back(vertex);
                            vertex.clear();
                        }
                    } // end if(real_segments[j].size() > 2)
                    else
                    {
                        frame.push_back(vertex);
                        frame.push_back(vertex);
                    }
                    if(frame.find_rgb_in_palette(p_space->palette_picker(LASERBOY_DXF)))
                        frame.palette_index = LASERBOY_DXF;
                    frame.is_2D();
                    push_back(frame);
                    frame.clear();
                }
                //------------------------------------------------------------
                renumber();
            } // end if(valid_data)
            else
            {
                frame.push_back(vertex);
                frame.push_back(vertex);
                frame.palette_index = LASERBOY_DXF;
                push_back(frame);
            }
        //--------------------------------------------------------------------
        } // end if(file_list.size())
        else
            error = LASERBOY_FILE_OPEN_FAILED;
    }
    if(error)
        return false;
    return true;
}

//############################################################################
bool LaserBoy_frame_set::from_LaserBoy_wave_file(string file, bool append)
{
    ifstream in(file.c_str(), ios::in | ios::binary);
    if(in.is_open())
    {
        LaserBoy_wave_header header(in);
        //--------------------------------------------------------------------
        if(      header.version == "!LaserBoy!"
            ||   header.num_channels < 6
            || !(header.LaserBoy_wave_mode & LASERBOY_WAVE_END_OF_FRAME)
          )
        {
            frame_index = 0;
            error       = from_ifstream_unframed_wave(header, in, append);
        }
        else
        {
            frame_index = 0;
            error       = from_ifstream_LaserBoy_wave(header, in, append);
        }
        in.close();
    }
    else
        error = LASERBOY_FILE_OPEN_FAILED;

    if(error)
        return false;
    return true;
}

//############################################################################
bool LaserBoy_frame_set::from_unframed_wave_file(string file, bool append)
{
    ifstream in(file.c_str(), ios::in | ios::binary);
    if(in.is_open())
    {
        LaserBoy_wave_header header(in);
        frame_index = 0;
        error       = from_ifstream_unframed_wave(header, in, append);
        in.close();
    }
    else
        error = LASERBOY_FILE_OPEN_FAILED;

    if(error)
        return false;
    return true;
}

//############################################################################
bool LaserBoy_frame_set::from_ADAT_wave_files(string file_1,
                                              string file_2,
                                              string file_3,
                                              string file_4
                                             )
{
    ifstream in_1,
             in_2,
             in_3,
             in_4;

    in_1.open(file_1.c_str(), ios::in | ios::binary);
    in_2.open(file_2.c_str(), ios::in | ios::binary);
    in_3.open(file_3.c_str(), ios::in | ios::binary);
    in_4.open(file_4.c_str(), ios::in | ios::binary);

    if(    in_1.is_open()
        && in_2.is_open()
        && in_3.is_open()
        && in_4.is_open()
      )
    {
        //error       = measure_ADAT_waves(in_1, in_2, in_3);
        //in_1.seekg(0, ios::beg);
        //in_2.seekg(0, ios::beg);
        //in_3.seekg(0, ios::beg);

        frame_index = 0;
        error       = from_ifstreams_ADAT_waves(in_1, in_2, in_3, in_4);
        in_1.close();
        in_2.close();
        in_3.close();
        in_4.close();
    }
    else
        error = LASERBOY_FILE_OPEN_FAILED;

    if(error)
        return false;
    return true;
}

//############################################################################
int LaserBoy_frame_set::from_ifstream_ild(ifstream& in, bool append)
{
    LaserBoy_header  header                    ;
    LaserBoy_frame   frame   (p_space)         ;
    LaserBoy_palette palette (p_space)         ;
    bool             true_color_data    = false,
                     some_data_accepted = false;
    //------------------------------------------------------------------------
    if(!append)
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    p_space->palette_index = LASERBOY_ILDA_DEFAULT;
    //------------------------------------------------------------------------
    while(header.from_ifstream_ild(in) == LASERBOY_OK)
    {
        if(header.quantity)
        {
            switch(header.format)
            {
                //------------------------------------------------------------
                case LASERBOY_3D_VECTOR:
                case LASERBOY_2D_VECTOR:
                    error |= frame.from_ifstream_ild(in, header);
                    if(!error)
                    {
                        frame.palette_index = p_space->palette_index;
                        frame.set_rgb_from_palette();
                        frame.is_2D() ? num_2D_frames++ : num_3D_frames++;
                        push_back(frame);
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
                case LASERBOY_PALETTE:
                    error |= palette.from_ifstream_ild(in, header);
                    if(!error)
                    {
                        bool built_in_palette = false;
                        for(int i = 0; i < LASERBOY_BUILT_IN_PALETTES; i++)
                            if(header.name == p_space->palette_picker(i).name)
                            {
                                p_space->palette_index = i;
                                built_in_palette = true;
                                break;
                            }
                        if(!built_in_palette)
                        {
                            p_space->push_back_palette(palette);
                            p_space->palette_index = p_space->number_of_palettes() - 1;
                        }
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
                case LASERBOY_COLOR_TABLE:
                    error |= frame.from_ifstream_format_3(in, header);
                    if(!error)
                    {
                        true_color_data = true;
                        frame.is_2D() ? num_2D_frames++ : num_3D_frames++;
                        push_back(frame);
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
                case LASERBOY_3D_COLOR_VECTOR:
                case LASERBOY_2D_COLOR_VECTOR:
                    error |= frame.from_ifstream_ild(in, header);
                    if(!error)
                    {
                        frame.palette_index = LASERBOY_TRUE_COLOR;
                        frame.is_2D() ? num_2D_frames++ : num_3D_frames++;
                        push_back(frame);
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
            }
        }
    }
    //------------------------------------------------------------------------
    if(!in.eof() && error && some_data_accepted)
        error = LASERBOY_OK;
    //------------------------------------------------------------------------
    if(true_color_data)
        p_space->palette_index = LASERBOY_TRUE_COLOR;
    //------------------------------------------------------------------------
    return error;
}
//############################################################################
int LaserBoy_frame_set::from_ifstream_LaserBoy_wave(LaserBoy_wave_header header,
                                                    ifstream& in,
                                                    bool append
                                                   )
{
    int                     i,
                            span,
                            last,
                            red,
                            green,
                            blue,
                            offset_max,
                            offset_min;

    LaserBoy_frame          frame(p_space);
    LaserBoy_vertex         vertex        ;
    LaserBoy_wave_sample_pointer  roll_over    ;
    LaserBoy_wave_sample_pointer *sample_window;
    //--------------------------------------------------------------------
    offset_min = header.offset[0];
    offset_max = header.offset[0];
    for(i = 1; i < 6; i++)
    {
        if(header.offset[i] > offset_max)
            offset_max = header.offset[i];
        if(header.offset[i] < offset_min)
            offset_min = header.offset[i];
    }
    span  = (offset_max - offset_min) + 2; // size of sample_window
    last  = span - 1                     ; // last index of sample_window
    red   = -header.offset[2] + 1        ; // index of red
    green = -header.offset[3] + 1        ; // index of green
    blue  = -header.offset[4] + 1        ; // index of blue
    //--------------------------------------------------------------------
    sample_window = new LaserBoy_wave_sample_pointer[span];
    for(i = 0; i < span; i++)
        sample_window[i] = new LaserBoy_wave_sample(header.num_channels);
    //--------------------------------------------------------------------
    frame.format        = LASERBOY_2D_VECTOR;
    frame.palette_index = LASERBOY_TRUE_COLOR;
    //--------------------------------------------------------------------
    if(!append) // to the current frame set
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    //--------------------------------------------------------------------
    while(true)
    {
        roll_over = sample_window[0];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;

        if(!sample_window[last]->from_ifstream_wave(in)) // if EOF
            break;

        if((header.LaserBoy_wave_mode & LASERBOY_WAVE_POSITIVE) == 0)
            sample_window[last]->negate();
        //----------------------------------------------------------------
        vertex.x = sample_window[1]->channel[0];
        vertex.y = sample_window[1]->channel[1];
        if(header.signal_id[5] == LASERBOY_SIGNAL_Z_POSITION)
            vertex.z = sample_window[1]->channel[5];
        else
            vertex.z = 0;
        vertex.r = (u_char)(sample_window[red  ]->channel[2] >> 7);
        vertex.g = (u_char)(sample_window[green]->channel[3] >> 7);
        vertex.b = (u_char)(sample_window[blue ]->channel[4] >> 7);
        if(vertex.r + vertex.g + vertex.b)
            vertex.k = 0;
        else
            vertex.k = LASERBOY_BLANKING_BIT;
        vertex.c = 0;
        //----------------------------------------------------------------
        frame.push_back(vertex);
        if(sample_window[red]->channel[2] & 0x0001) // end of frame
        {
            frame.is_unique = true;
            if(    (header.LaserBoy_wave_mode & LASERBOY_WAVE_UNIQUE_FRAME)  // if we are looking for it
                && (!(sample_window[green]->channel[3] & 0x0001)) // and we do not find it
              )
                frame.is_unique = false;
            frame.back().blank();
            frame.quantity = (u_short)frame.number_of_vertices();
            push_back(frame);
            frame.clear();
        }
    }
    //--------------------------------------------------------------------
    renumber();
    //--------------------------------------------------------------------
    delete sample_window;
    error = LASERBOY_OK;
    return error;
}

//############################################################################
int LaserBoy_frame_set::from_ifstream_unframed_wave(  LaserBoy_wave_header header
                                                     ,ifstream&            in
                                                     , bool                append
                                                   )
{
    int                     i,
                            vertices = 0,
                            span,
                            last,
                            red,
                            green,
                            blue,
                            offset_max,
                            offset_min;

    LaserBoy_frame          frame(p_space);
    LaserBoy_vertex         vertex;
    LaserBoy_wave_sample_pointer  roll_over;
    LaserBoy_wave_sample_pointer *sample_window;
    //--------------------------------------------------------------------
    offset_max = p_space->wave_offsets[0];
    for(i = 1; i < 6; i++)
        if(p_space->wave_offsets[i] > offset_max)
            offset_max = p_space->wave_offsets[i];
    //--------------------------------------------------------------------
    for(i = 0; i < 6; i++)
        p_space->wave_offsets[i] -= offset_max;
    //--------------------------------------------------------------------
    for(i = 0; i < 6; i++)
        header.offset[i] = p_space->wave_offsets[i]; // use current system offsets
    //--------------------------------------------------------------------
    offset_min = header.offset[0];
    offset_max = header.offset[0];
    for(i = 1; i < 6; i++)
    {
        if(header.offset[i] > offset_max)
            offset_max = header.offset[i];
        if(header.offset[i] < offset_min)
            offset_min = header.offset[i];
    }
    //--------------------------------------------------------------------
    span  = (offset_max - offset_min) + 2; // at least 2 samples!
    last  = span - 1;
    red   = -header.offset[2] + 1;
    green = -header.offset[3] + 1;
    blue  = -header.offset[4] + 1;
    //------------------------------------------------------------------------
    sample_window = new LaserBoy_wave_sample_pointer[span];
    for(i = 0; i < span; i++)
        sample_window[i] = new LaserBoy_wave_sample(header.num_channels);
    //------------------------------------------------------------------------
    frame.format        = LASERBOY_2D_VECTOR;
    frame.palette_index = LASERBOY_TRUE_COLOR;
    //------------------------------------------------------------------------
    if(!append) // to the current frame set
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    //------------------------------------------------------------------------
    while(true)
    {
        roll_over = sample_window[0];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;

        if(!sample_window[last]->from_ifstream_wave(in)) // if EOF
            break;

        vertices++;
        vertex.x = sample_window[1]->channel[0];
        vertex.y = sample_window[1]->channel[1];
        vertex.z = 0;
        switch(header.num_channels)
        {
            case 2: vertex.r = 255;
                    vertex.g = 255;
                    vertex.b = 255;
                    break;
            //----------------------------------------------------------------
            case 4: vertex.r =
                    vertex.g =
                    vertex.b = (u_char) abs(sample_window[red  ]->channel[2] / 128);
                    break;
            //----------------------------------------------------------------
            case 6:
            case 8: vertex.r = (u_char) abs(sample_window[red  ]->channel[2] / 128);
                    vertex.g = (u_char) abs(sample_window[green]->channel[3] / 128);
                    vertex.b = (u_char) abs(sample_window[blue ]->channel[4] / 128);
                    break;
            //----------------------------------------------------------------
        }
        vertex.k = 0;
        vertex.c = 0;
        //--------------------------------------------------------------------
        frame.push_back(vertex);

        if(!(vertices % p_space->unframed_wave_stride))
        {
            frame.is_unique = true;
            frame.quantity = (u_short)frame.number_of_vertices();
            push_back(frame);
            frame.clear();
        }
    }
    //------------------------------------------------------------------------
    frame.is_unique = true;
    frame.quantity = (u_short)frame.number_of_vertices();
    push_back(frame);
    frame.clear();
    renumber();
    error = LASERBOY_OK;
    return error;
}

//############################################################################
int LaserBoy_frame_set::measure_ADAT_waves(//ifstream& in_1,
                                          ifstream& in_2,
                                          ifstream& in_3,
                                          ifstream& in_4
                                         )
{
    LaserBoy_wave_header //header_1,
                   header_2(in_2),
                   header_3(in_3),
                   header_4(in_4);
    LaserBoy_wave_sample //sample_1(2),
                   sample_2(1),
                   sample_3(1),
                   sample_4(1);
    //------------------------------------------------------------------------
    if(    header_2.version == "!LaserBoy!"
        || header_3.version == "!LaserBoy!"
        || header_4.version == "!LaserBoy!"
      )
        return LASERBOY_BAD_HEADER;
    //------------------------------------------------------------------------
    max_r = LASERBOY_MIN_SHORT;
    max_g = LASERBOY_MIN_SHORT;
    max_b = LASERBOY_MIN_SHORT;

    while(    sample_2.from_ifstream_wave_inverted(in_2)
           && sample_3.from_ifstream_wave_inverted(in_3)
           && sample_4.from_ifstream_wave_inverted(in_4)
         )
    {
        if(max_r < sample_2.channel[0])
            max_r = sample_2.channel[0];

        if(max_g < sample_3.channel[0])
            max_g = sample_3.channel[0];

        if(max_b < sample_4.channel[0])
            max_b = sample_4.channel[0];
    }
    //------------------------------------------------------------------------
    return LASERBOY_OK;
}

//############################################################################
int LaserBoy_frame_set::from_ifstreams_ADAT_waves(ifstream& in_1,
                                                  ifstream& in_2,
                                                  ifstream& in_3,
                                                  ifstream& in_4
                                                 )
{
    int             i,
                    color_sum,
                    static_sample_count = 0;

    LaserBoy_frame  frame(p_space);

    LaserBoy_vertex vertex,
                    vertex_1,
                    vertex_2,
                    vertex_3,
                    vertex_4;

    LaserBoy_wave_header header_1(in_1),
                         header_2(in_2),
                         header_3(in_3),
                         header_4(in_4);

    LaserBoy_wave_sample sample_1(2),
                         sample_2(1),
                         sample_3(1),
                         sample_4(1);

//    double         red_gain   = (LASERBOY_MAX_SHORT / max_r) / 128.0,
//                   green_gain = (LASERBOY_MAX_SHORT / max_g) / 128.0,
//                   blue_gain  = (LASERBOY_MAX_SHORT / max_b) / 128.0;

    double         red_gain   = 1.0 / 128.0,
                   green_gain = 1.0 / 128.0,
                   blue_gain  = 1.0 / 128.0;
    //------------------------------------------------------------------------
    frame.format        = LASERBOY_2D_VECTOR;
    frame.palette_index = LASERBOY_TRUE_COLOR;
    num_2D_frames       = 0;
    num_3D_frames       = 0;
    clear();
    //------------------------------------------------------------------------
    for(i = 0; i < 7; i++) // color time alignment woo hoo!
    {
        sample_2.from_ifstream_wave_inverted(in_2); // eat-um'up!
        sample_3.from_ifstream_wave_inverted(in_3); // eat-um'up!
        sample_4.from_ifstream_wave_inverted(in_4); // eat-um'up!
    }
    //------------------------------------------------------------------------
    while(    sample_1.from_ifstream_wave_inverted(in_1)
           && sample_2.from_ifstream_wave_inverted(in_2)
           && sample_3.from_ifstream_wave_inverted(in_3)
           && sample_4.from_ifstream_wave_inverted(in_4)
         )
    {
        vertex.x = sample_1.channel[0];
        vertex.y = sample_1.channel[1];
        vertex.z = 0;

        color_sum =   sample_2.channel[0]
                    + sample_3.channel[0]
                    + sample_4.channel[0];

        if(color_sum < 400) // very close to black!
        {
            vertex.r = 255; // if unblanked it is red
            vertex.g = 0;
            vertex.b = 0;
            vertex.k = LASERBOY_BLANKING_BIT; // blank it
            vertex.c = 0; // 24 bit

            if(    (vertex | vertex_2) <= p_space->insignificant_distance
                && (vertex | vertex_4) <= p_space->insignificant_distance
              )
                static_sample_count++;

            if(static_sample_count > 150) // we found a frame boundary
            {
                frame.is_unique = true;
                frame.quantity = (u_short)frame.number_of_vertices();
                if(frame.quantity > 1)
                {
                    frame.at(1).blank();
                    push_back(frame);
                    frame.clear();
                }
                static_sample_count = 0;

                while(    sample_1.from_ifstream_wave_inverted(in_1)
                       && sample_2.from_ifstream_wave_inverted(in_2)
                       && sample_3.from_ifstream_wave_inverted(in_3)
                       && sample_4.from_ifstream_wave_inverted(in_4)
                     )
                {
                    color_sum =   sample_2.channel[0]
                                + sample_3.channel[0]
                                + sample_4.channel[0];
                    if(color_sum > 400) // we have color!
                        break;
                }
            }
            vertex_4 = vertex_3;
            vertex_3 = vertex_2;
            vertex_2 = vertex_1;
            vertex_1 = vertex;
        }
        else
        {
            vertex.r = (u_char)((sample_2.channel[0] > 0) ? (sample_2.channel[0] *   red_gain) : (0));
            vertex.g = (u_char)((sample_3.channel[0] > 0) ? (sample_3.channel[0] * green_gain) : (0));
            vertex.b = (u_char)((sample_4.channel[0] > 0) ? (sample_4.channel[0] *  blue_gain) : (0));
            vertex.k = 0;
            vertex.c = 0;
            static_sample_count = 0;
        }
        //--------------------------------------------------------------------
        frame.push_back(vertex);
        //--------------------------------------------------------------------
    }
    renumber();
    //------------------------------------------------------------------------
    return LASERBOY_OK;
}

//############################################################################
int LaserBoy_frame_set::from_ifstream_dxf(ifstream& in, bool append)
{
    LaserBoy_frame frame(p_space);
    //------------------------------------------------------------------------
    if(!append)
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    //------------------------------------------------------------------------
    frame.from_ifstream_dxf(in);
    //------------------------------------------------------------------------
    p_space->palette_index = frame.palette_index;
    push_back(frame);
    renumber();
    //------------------------------------------------------------------------
    return error;
}

//############################################################################
bool LaserBoy_frame_set::is_2D()
{
    for(int i = 0; i < number_of_frames(); i++)
        if(!at(i).is_2D())
            return false;
    return true;
}

//############################################################################
void LaserBoy_frame_set::flatten_z()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).flatten_z();
    return;
}

//############################################################################
void LaserBoy_frame_set::z_order_vertices(unsigned short span)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).z_order_vertices(span);
    return;
}

//############################################################################
void LaserBoy_frame_set::unblank_all_vertices()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).unblank_all_vertices();
    return;
}

//############################################################################
void LaserBoy_frame_set::convert_black_to_blank()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).convert_black_to_blank();
    return;
}

//############################################################################
void LaserBoy_frame_set::reduce_blank_vectors()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).reduce_blank_vectors();
    return;
}

//############################################################################
void LaserBoy_frame_set::remove_dots()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).remove_dots();
    return;
}

//############################################################################
void LaserBoy_frame_set::enhance_dots()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).enhance_dots();
    return;
}

//############################################################################
void LaserBoy_frame_set::remove_dwell_vertices()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).remove_dwell_vertices();
    return;
}

//############################################################################
void LaserBoy_frame_set::reduce_lit_vectors()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).reduce_lit_vectors();
    return;
}

//############################################################################
void LaserBoy_frame_set::minimize()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).minimize();
    return;
}

//############################################################################
void LaserBoy_frame_set::to_dots()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).to_dots();
    return;
}

//############################################################################
void LaserBoy_frame_set::accentuate_the_positive()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).accentuate_the_positive();
    return;
}

//############################################################################
void LaserBoy_frame_set::eliminate_the_negative()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).eliminate_the_negative();
    return;
}

//############################################################################
void LaserBoy_frame_set::impose_bit_resolution()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).impose_bit_resolution();
    return;
}

//############################################################################
void LaserBoy_frame_set::explode_current_frame()
{
    int                 i;
    LaserBoy_frame_set  results(p_space),
                        exploded = current_frame().find_segments();

    for(i = 0; i < frame_index; i++)
        results += at(i);

    results += exploded;

    for(i = frame_index + 1; i < number_of_frames(); i++)
        results += at(i);

    *this = results;
    renumber();
    return;
}

//############################################################################
void LaserBoy_frame_set::explode_all_frames()
{
    int                 i;
    LaserBoy_frame_set  results (p_space),
                        exploded(p_space);

    for(i = 0; i < number_of_frames(); i++)
        results +=  at(i).find_segments();

    *this = results;
    renumber();
    return;
}

//############################################################################
void LaserBoy_frame_set::collapse_selected_frames()
{
    LaserBoy_frame  collapsed(p_space);
    int             first_selected_frame_index = -1;

    for(int i = 0; i < number_of_frames(); i++)
    {
        if(at(i).is_selected)
        {
            if(first_selected_frame_index == -1)
            {
                first_selected_frame_index = i;
                collapsed.palette_index = at(i).palette_index;
            }
            collapsed += at(i);
        }
    }
    if(first_selected_frame_index != -1)
    {
        frame(first_selected_frame_index) = collapsed;
        frame(first_selected_frame_index).is_selected = false;
        delete_selected_frames();
        renumber();
    }
    return;
}

//############################################################################
void LaserBoy_frame_set::delete_every_nth_frame(unsigned short n)
{
    LaserBoy_frame_set results(p_space);
    if(n > 1 && n < number_of_frames())
    {
        for(int i = 0; i < number_of_frames(); i++)
            if(i % n)
                results += at(i);
        *this = results;
        renumber();
    }
    return;
}

//############################################################################
void LaserBoy_frame_set::fracture_segments()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).fracture_segments();
    return;
}

//############################################################################
void LaserBoy_frame_set::blank_equivalent_vectors()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).blank_equivalent_vectors();
    return;
}

//############################################################################
void LaserBoy_frame_set::bond_segments()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).bond_segments();
    return;
}

//############################################################################
void LaserBoy_frame_set::reorder_segments()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).reorder_segments();
    return;
}

//############################################################################
void LaserBoy_frame_set::add_dwell_vertices()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).add_dwell_vertices();
    return;
}

//############################################################################
void LaserBoy_frame_set::add_lit_span_vertices()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).add_lit_span_vertices();
    return;
}

//############################################################################
void LaserBoy_frame_set::add_blank_span_vertices()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).add_blank_span_vertices();
    return;
}

//############################################################################
void LaserBoy_frame_set::delete_redundant_frames()
{
    LaserBoy_frame_set reduced(p_space);
    for(int i = 0; i < number_of_frames(); i++)
        if(at(i).is_unique)
            reduced.push_back(at(i));
    reduced.renumber();
    *this = reduced;
    return;
}

//############################################################################
void LaserBoy_frame_set::optimize()
{
    if(number_of_frames())
    {
        int i,
            dwell_vertex_count = 0,
            dwell_vertex_total = 0,
            blank_dwell       = p_space->sample_rate / 220;
        //--------------------------------------------------------------------
        for(i = 0; i < number_of_frames(); i++)
            at(i).optimize();
        //--------------------------------------------------------------------
        intro.clear();
        for(i = 0; i < blank_dwell; i++)
            intro.push_back(LaserBoy_vertex()); // origin
        //--------------------------------------------------------------------
        intro += LaserBoy_segment(p_space, intro[0], front()[0]); // from origin to first vertex in first frame
        //--------------------------------------------------------------------
        for(i = 0; i < blank_dwell; i++)
            intro.push_back(front()[0]);
        //--------------------------------------------------------------------
        for(i = 0; i < (number_of_frames() - 1); i++)
        {
            at(i).transition_to_next_frame.clear();
            dwell_vertex_total = dwell_vertices(delta_angle(at(i)[at(i).number_of_vertices() - 2] | at(i + 1).front(),
                                                           at(i)[at(i).number_of_vertices() - 2] | at(i).back(),
                                                           at(i).back()                        | at(i + 1).front()
                                                          ),
                                               p_space->sample_rate,
                                               p_space->max_dwell_microsec
                                              );
            if(dwell_vertex_total)
                for(dwell_vertex_count = 0; dwell_vertex_count < dwell_vertex_total; dwell_vertex_count++)
                    at(i).transition_to_next_frame.push_back(at(i).back());
            //----------------------------------------------------------------
            at(i).transition_to_next_frame += LaserBoy_segment(p_space, at(i).back(), at(i + 1)[0]);
        }
        //--------------------------------------------------------------------
        back().transition_to_next_frame = LaserBoy_segment(p_space, back().back(), intro[0]);
        for(i = 0; i < blank_dwell; i++)
            back().transition_to_next_frame.push_back(intro[0]);
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void LaserBoy_frame_set::rainbow_recolor(int effect)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).rainbow_recolor(effect);
    return;
}

//############################################################################
void LaserBoy_frame_set::quarter_turn(int plane, int turns)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).quarter_turn(plane, turns);
    return;
}

//############################################################################
int LaserBoy_frame_set::scale_around_origin(LaserBoy_3D_double factor)
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).scale_around_origin(factor);
    return all_in_bounds;
}

//############################################################################
int LaserBoy_frame_set::scale_around_origin()
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).scale_around_origin();
    return all_in_bounds;
}

//############################################################################
int LaserBoy_frame_set::rotate_around_origin(LaserBoy_3D_double angle)
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).rotate_around_origin(angle);
    return all_in_bounds;
}

//############################################################################
int LaserBoy_frame_set::rotate_around_origin()
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).rotate_around_origin();
    return all_in_bounds;
}

//############################################################################
int LaserBoy_frame_set::move(LaserBoy_3D_double offset)
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).move(offset);
    return all_in_bounds;
}

//############################################################################
int LaserBoy_frame_set::move()
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).move();
    return all_in_bounds;
}

//############################################################################
void LaserBoy_frame_set::toggle_frame_select_start()
{
    int i;
    if(frame_select_start == -1) // was unset
    {
        frame_select_start = frame_index; // set
        if(frame_select_end != -1)
        {
            for(i = frame_select_start;
                i != frame_select_end;
                (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = true;
            at(frame_select_end).is_selected = true;
        }
    }
    else // was set
    {
        if(frame_select_end != -1) // was set
        {
            for(i = frame_select_start;
                i != frame_select_end;
                (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = false;
            at(frame_select_end).is_selected = false;
        }
        frame_select_start = -1; // unset
    }
    return;
}

//############################################################################
void LaserBoy_frame_set::toggle_frame_select_end()
{
    int i;
    if(frame_select_end == -1) // was unset
    {
        frame_select_end = frame_index; // set
        if(frame_select_start != -1)
        {
            for(i = frame_select_start;
                i != frame_select_end;
                (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = true;
            at(frame_select_end).is_selected = true;
        }
    }
    else // was set
    {
        if(frame_select_start != -1) // was set
        {
            for(i = frame_select_start;
                i != frame_select_end;
                (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = false;
            at(frame_select_end).is_selected = false;
        }
        frame_select_end = -1; // unset
    }
    return;
}

//############################################################################
void LaserBoy_frame_set::delete_selected_frames()
{
    LaserBoy_frame_set reduced(p_space);
    for(int i = 0; i < number_of_frames(); i++)
    {
        if(!at(i).is_selected)
            reduced.push_back(at(i));
    }
    reduced.renumber();

    if(    frame_select_start != -1
        && frame_select_end   != -1
      )
    {
        frame_index = (frame_select_start - 1);
        frame_select_start = -1;
        frame_select_end   = -1;
    }

    if(frame_index < 0)
        frame_index = 0;

    if(frame_index > reduced.number_of_frames())
        frame_index = reduced.number_of_frames() - 1;

    *this = reduced;
    return;
}

//############################################################################
void LaserBoy_frame_set::trim_to_selected_frames()
{
    LaserBoy_frame_set reduced(p_space);
    reduced.reserve(size());
    for(int i = 0; i < number_of_frames(); i++)
    {
        if(at(i).is_selected)
            reduced.push_back(at(i));
    }
    reduced.renumber();
    if(    frame_select_start != -1
        && frame_select_end   != -1
      )
    {
        frame_select_start = -1;
        frame_select_end   = -1;
    }
    frame_index = 0;
    *this = reduced;
    return;
}

//############################################################################
void LaserBoy_frame_set::bit_reduce_to_palette()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).bit_reduce_to_palette();
    return;
}

//############################################################################
void LaserBoy_frame_set::best_reduce_to_palette()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).best_reduce_to_palette();
    return;
}

//############################################################################
void LaserBoy_frame_set::promote_to_true_color()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).promote_to_true_color();
    return;
}

//############################################################################
void LaserBoy_frame_set::best_match_palette(int index)
{
    if(!p_space->allow_lit_black)
        convert_black_to_blank();
    for(int i = 0; i < number_of_frames(); i++)
        at(i).best_match_palette(index);
    return;
}

//############################################################################
void LaserBoy_frame_set::best_match_target_palette()
{
    if(!p_space->allow_lit_black)
        convert_black_to_blank();
    for(int i = 0; i < number_of_frames(); i++)
        at(i).best_match_target_palette();
    return;
}

//############################################################################
void LaserBoy_frame_set::strip_color()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).strip_color();
    return;
}

//############################################################################
void LaserBoy_frame_set::strip_color_or()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).strip_color_or();
    return;
}

//############################################################################
void LaserBoy_frame_set::strip_color_avg()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).strip_color_avg();
    return;
}

//############################################################################
void LaserBoy_frame_set::accept_palette(int index)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).accept_palette(index);
    return;
}

//############################################################################
void LaserBoy_frame_set::accept_target_palette()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).accept_target_palette();
    return;
}

//############################################################################
bool LaserBoy_frame_set::color_from_bmp(string file)
{
    char file_name[256];
    struct LaserBoy_bmp bmp;

    strcpy(file_name, file.c_str());
    if(bmp_from_file(&bmp, file_name))
    {
        color_from_bmp(&bmp);
        bmp_free(&bmp);
        return true;
    }
    return false;
}

//############################################################################
void LaserBoy_frame_set::color_from_bmp(struct LaserBoy_bmp* bmp)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).color_from_bmp(bmp);
    return;
}

//############################################################################
bool LaserBoy_frame_set::subtract_bmp(string file)
{
    char file_name[81];
    struct LaserBoy_bmp bmp;

    strcpy(file_name, (file).c_str());
    if(bmp_from_file(&bmp, file_name))
    {
        subtract_bmp(&bmp);
        bmp_free(&bmp);
        return true;
    }
    return false;
}

//############################################################################
void LaserBoy_frame_set::subtract_bmp(struct LaserBoy_bmp* bmp)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).subtract_bmp(bmp);
    return;
}

//############################################################################
void LaserBoy_frame_set::flip(int plane)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).flip(plane);
    return;
}

//############################################################################
void LaserBoy_frame_set::renumber()
{
    int   i,
          local_palette_index = LASERBOY_ILDA_DEFAULT,
          added_palettes      = 0,
          frame_id            = 0;
    bool  in_true_color       = false;
    char  number[7];
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_frames(); i++)
    {
        if((at(i).quantity = at(i).number_of_vertices()) > 1)
        {
            if(!p_space->dumb_file_format)
            {
                if(at(i).palette_index == LASERBOY_TRUE_COLOR)
                {
                    if(!in_true_color)
                    {
                        frame_id++;
                        added_palettes++; // add LASERBOY_332
                        local_palette_index = LASERBOY_LASERBOY_332;
                        in_true_color = true;
                    }
                    frame_id++;
                    added_palettes++; // add true_color
                }
                else if(at(i).palette_index != local_palette_index)
                {
                    frame_id++;
                    added_palettes++;
                    local_palette_index = at(i).palette_index;
                    in_true_color = false;
                }
            }
            sprintf(number, "%06d", frame_id);
            at(i).identity = frame_id;
            at(i).name     = "ES" + string(number);
            frame_id++;
        }
        else
            delete_frame(i);
    }
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_frames(); i++)
        at(i).total = (u_short)(number_of_frames() + added_palettes);
    if(frame_index >= number_of_frames())
        frame_index = number_of_frames() - 1;
    return;
}

//############################################################################
LaserBoy_frame_set& LaserBoy_frame_set::reverse()
{
    LaserBoy_frame_set reversed(p_space);
    for(int i = 1; i <= number_of_frames(); i++)
        reversed.push_back(at(number_of_frames() - i));
    clear();
    insert(begin(), reversed.begin(), reversed.end());
    renumber();
    return *this;
}

//############################################################################
LaserBoy_frame_set& LaserBoy_frame_set::make_current_first()
{
    int                i,
                       j;
    LaserBoy_frame_set new_order(p_space);
    //------------------------------------------------------------------------
    for(i = frame_index; i < number_of_frames(); i++)
        new_order += at(i);
    //------------------------------------------------------------------------
    for(i = 0; i < frame_index; i++)
        new_order += at(i);
    //------------------------------------------------------------------------
    *this = new_order;
    frame_index = 0;
    //------------------------------------------------------------------------
    return *this;
}

//############################################################################
LaserBoy_frame_set& LaserBoy_frame_set::reverse_selected()
{
    int                i,
                       j;
    LaserBoy_frame_set selected(p_space);
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_frames(); i++)
        if(at(i).is_selected)
            selected += at(i);
    //------------------------------------------------------------------------
    j = selected.number_of_frames();
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_frames(); i++)
        if(at(i).is_selected)
            at(i) = selected[--j];
    //------------------------------------------------------------------------
    return *this;
}

//############################################################################
LaserBoy_frame_set& LaserBoy_frame_set::random_order()
{
    LaserBoy_frame_set random(p_space);
    vector<bool>       been_here(number_of_frames());
    //------------------------------------------------------------------------
    int random_index = rand() % number_of_frames();
    //------------------------------------------------------------------------
    for(int i = 0; i < number_of_frames(); i++)
        been_here[i] = false;
    //------------------------------------------------------------------------
    for(int i = 0; i < number_of_frames(); i++)
    {
        while(been_here[random_index])
        {
            random_index = rand() % number_of_frames();
        }
        random.push_back(at(random_index));
        been_here[random_index] = true;
    }
    //------------------------------------------------------------------------
    clear();
    insert(begin(), random.begin(), random.end());
    renumber();
    return *this;
}

//############################################################################
LaserBoy_frame_set& LaserBoy_frame_set::frame_reverse()
{
    for(int i = 1; i <= number_of_frames(); i++)
        at(i).reverse();
    return *this;
}

//############################################################################
bool LaserBoy_frame_set::save_as_ild(string file)
{
    ofstream out(file.c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        to_ofstream_ild(out);
        out.close();
    }
    else
        return false;
    return true;
}

//############################################################################
void LaserBoy_frame_set::to_ofstream_ild(ofstream& out)
{
    int             file_section_number = 0,
                    local_palette_index = LASERBOY_ILDA_DEFAULT;
    bool            in_true_color       = false;
    LaserBoy_header  header;
    //------------------------------------------------------------------------
    renumber();
    header.format  = LASERBOY_PALETTE;
    header.total   = front().total;
    //------------------------------------------------------------------------
    for(int i = 0; i < number_of_frames(); i++)
    {
        if(!p_space->dumb_file_format)
        {
            if(at(i).palette_index == LASERBOY_TRUE_COLOR)
            {
                if(!in_true_color)
                {
                    header.name         = p_space->palette_picker(LASERBOY_LASERBOY_332).name;
                    header.quantity     = p_space->palette_picker(LASERBOY_LASERBOY_332).number_of_colors();
                    header.identity     = file_section_number;
                    p_space->palette_picker(LASERBOY_LASERBOY_332).to_ofstream_ild(out, header);
                    local_palette_index = LASERBOY_LASERBOY_332;
                    file_section_number++;
                    in_true_color = true;
                }
                file_section_number++;
            }
            else if(at(i).palette_index != local_palette_index)
            {
                header.name         = p_space->palette_picker(at(i).palette_index).name;
                header.quantity     = p_space->palette_picker(at(i).palette_index).number_of_colors();
                header.identity     = file_section_number;
                p_space->palette_picker(at(i).palette_index).to_ofstream_ild(out, header);
                local_palette_index = at(i).palette_index;
                file_section_number++;
                in_true_color = false;
            }
        }
        at(i).is_2D(); // checks for z and sets format accordingly;
        at(i).to_ofstream_ild(out);
        file_section_number++;
    }
    //------------------------------------------------------------------------
    header.format   = LASERBOY_3D_VECTOR  ;
    header.name     = "ILDA_End"         ;
    header.quantity = 0                  ;
    header.identity = file_section_number;
    header.to_ofstream_ild(out)         ;
    return;
}

//############################################################################
bool LaserBoy_frame_set::save_as_wave(string file)
{
    ofstream out(file.c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        if(number_of_frames() > 1)
            to_ofstream_wave(out);
        else
        {
            LaserBoy_frame copy(front());
            copy.rotate_around_origin();
            if(p_space->auto_flatten_z)
                copy.flatten_z();
            if(p_space->auto_optimize)
                copy.optimize();
            copy.save_as_wave(file);
        }
        out.close();
        return true;
    }
    return false;
}

//############################################################################
void LaserBoy_frame_set::to_ofstream_wave(ofstream& out)
{
    int                   i;
    LaserBoy_wave_header  header(   p_space->sample_rate
                                  ,    LASERBOY_WAVE_POSITIVE
                                     | LASERBOY_WAVE_END_OF_FRAME
                                     | LASERBOY_WAVE_UNIQUE_FRAME
                                     | LASERBOY_WAVE_SIGNAL_MATRIX
                                  , 6
                                );
    //------------------------------------------------------------------------
    LaserBoy_frame_set  copy(*this);
    //------------------------------------------------------------------------
    header.signal_id [0] = LASERBOY_SIGNAL_X_POSITION  ; // default values
    header.signal_id [1] = LASERBOY_SIGNAL_Y_POSITION  ;
    header.signal_id [2] = LASERBOY_SIGNAL_RED_ANALOG  ;
    header.signal_id [3] = LASERBOY_SIGNAL_GREEN_ANALOG;
    header.signal_id [4] = LASERBOY_SIGNAL_BLUE_ANALOG ;
    header.signal_id [5] = LASERBOY_SIGNAL_Z_POSITION  ;

    header.LSB_tag   [0] = LASERBOY_LSB_NOT_USED    ; // default values
    header.LSB_tag   [1] = LASERBOY_LSB_NOT_USED    ;
    header.LSB_tag   [2] = LASERBOY_LSB_END_OF_FRAME;
    header.LSB_tag   [3] = LASERBOY_LSB_UNIQUE_FRAME;
    header.LSB_tag   [4] = LASERBOY_LSB_NOT_USED    ;
    header.LSB_tag   [5] = LASERBOY_LSB_NOT_USED    ;

    if(    p_space->signal_bit_mask[0] != 0
        || p_space->signal_bit_mask[1] != 0
        || p_space->signal_bit_mask[2]  > 7
        || p_space->signal_bit_mask[3]  > 7
        || p_space->signal_bit_mask[4]  > 7
        || p_space->signal_bit_mask[5] != 0
      )
    {
        header.resolution[0] = 16 - p_space->signal_bit_mask[0];
        header.resolution[1] = 16 - p_space->signal_bit_mask[1];
        header.resolution[2] = 16 - p_space->signal_bit_mask[2];
        header.resolution[3] = 16 - p_space->signal_bit_mask[3];
        header.resolution[4] = 16 - p_space->signal_bit_mask[4];
        header.resolution[5] = 16 - p_space->signal_bit_mask[5];
        header.LaserBoy_wave_mode |= LASERBOY_SIGNAL_BIT_RESOLUTION;
    }
    //------------------------------------------------------------------------
    for(i = 0; i < copy.number_of_frames(); i++)
        copy[i].rotate_around_origin();
    //------------------------------------------------------------------------
    if(is_2D() || p_space->auto_flatten_z)
    {
        header.signal_id [5] = p_space->signal_if_not_z;
        header.resolution[5] = 8;
        copy.flatten_z();
    }
    //------------------------------------------------------------------------
    if(p_space->auto_optimize)
    {
        header.LaserBoy_wave_mode |= LASERBOY_WAVE_OPTIMIZED;
        header.parms.lit_dwell_overhang     =        p_space->lit_dwell_overhang    ;
        header.parms.lit_delta_max          = (float)p_space->lit_delta_max         ;
        header.parms.blank_delta_max        = (float)p_space->blank_delta_max       ;
        header.parms.max_dwell_microsec     = (float)p_space->max_dwell_microsec    ;
        header.parms.insignificant_distance = (float)p_space->insignificant_distance;
        header.parms.insignificant_angle    = (float)p_space->insignificant_angle   ;
        header.parms.frames_per_second      = (float)p_space->frames_per_second     ;
        copy.optimize();
    }
    //------------------------------------------------------------------------
    if(p_space->invert_wave_output)
        header.LaserBoy_wave_mode &= ~LASERBOY_WAVE_POSITIVE;
    //------------------------------------------------------------------------
    header.to_ofstream_wave(out); // take up the space!
    //------------------------------------------------------------------------
    if(p_space->auto_optimize)
        copy.intro.to_ofstream_wave(out, header);
    //------------------------------------------------------------------------
    for(i = 0; i < copy.number_of_frames(); i++)
        copy.at(i).to_ofstream_wave(out, header);
    //------------------------------------------------------------------------
    header.to_ofstream_wave(out);
    //------------------------------------------------------------------------
    return;
}

//############################################################################
bool LaserBoy_frame_set::save_as_bmp_directory(string dir) const
{
    char name[8];
#ifndef WIN32
    mkdir(dir.c_str(), 0700);
#else
    mkdir(dir.c_str());
#endif
    for(int i = 0; i < number_of_frames(); i++)
    {
        sprintf(name, "%06d", i);
        at(i).save_as_bmp(string(dir + '/' + name));
    }
    return true;
}

//############################################################################
bool LaserBoy_frame_set::save_as_dxf_directory(string dir) const
{
    bool all_saved = true;
    char name[8];
#ifndef WIN32
    mkdir(dir.c_str(), 0700);
#else
    mkdir(dir.c_str());
#endif
    for(int i = 0; i < number_of_frames(); i++)
    {
        sprintf(name, "%06d", i);
        all_saved &= at(i).save_as_dxf(string(dir + '/' + name));
    }
    return all_saved;
}

//############################################################################
bool LaserBoy_frame_set::save_as_code(string object_name, string dir)
{
    ofstream out((dir + object_name + ".cpp").c_str(), ios::out);
    if(out.is_open())
    {
        out << "//############################################################################\n"
               "// ~  This file is the product of LaserBoy!  ~\n"
               "//\n"
               "// by James Lehman\n"
               "// james@akrobiz.com\n"
               "//\n"
               "//############################################################################\n\n\n\n"
               "//############################################################################\n"
            << "LaserBoy_frame_set ";
        out << object_name;
        out << "()\n"
               "{\n"
               "    LaserBoy_frame     frame;\n"
               "    LaserBoy_frame_set frames;\n"
               "    int                frame_index;\n\n"
               "    for(frame_index = 0; frame_index < ";
        out << number_of_frames();
        out << "; frame_index++)\n"
               "    {\n";
        for(int i = 0; i < number_of_frames(); i++)
        {

            at(i).to_ofstream_code(out, "        ");
            out << "        frames.push_back(frame);\n"
                   "        frame.clear();\n\n";
        }
        out << "    }\n"
               "    return frames;\n"
               "}\n\n"
               "//############################################################################\n"
               "//////////////////////////////////////////////////////////////////////////////\n"
               "//############################################################################\n\n\n";

        out.close();
        return true;
    }
    return false;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
