//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
// Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009 James Lehman.
// This source is distributed under the terms of the GNU General Public License.
//
// LaserBoy_palettes.hpp is part of LaserBoy.
//
// LaserBoy is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// LaserBoy is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LaserBoy. If not, see <http://www.gnu.org/licenses/>.
//
//############################################################################
#ifndef __LASERBOY_PALETTE_DEFINITIONS__
#define __LASERBOY_PALETTE_DEFINITIONS__

//############################################################################
#include "LaserBoy_bmp.hpp"
#include "LaserBoy_color.hpp"
#include "LaserBoy_header.hpp"

//############################################################################
class LaserBoy_space;

//############################################################################
class LaserBoy_palette : public vector<LaserBoy_color>
{
public:
    LaserBoy_palette(LaserBoy_space* ps = NULL)
                    : p_space(ps)
                                                     {}
    //------------------------------------------------------------------------
    LaserBoy_palette(LaserBoy_space* ps, const int& size)
                    : p_space(ps)
                    {
                        reserve(size);
                    }
    //------------------------------------------------------------------------
    LaserBoy_palette(const LaserBoy_palette& p)
              : in_use  (true     ),
                white   (p.white  ),
                black   (p.black  ),
                index   (p.index  ),
                first   (p.first  ),
                last    (p.last   ),
                dud     (         ),
                name    (p.name   ),
                p_space (p.p_space)
                {
                    clear();
                    reserve(p.size());
                    insert(begin(), p.begin(), p.end());
                }
    //------------------------------------------------------------------------
   ~LaserBoy_palette()   {    clear();    }
    //------------------------------------------------------------------------
    LaserBoy_palette&   operator = (const LaserBoy_palette& p)
                    {
                        reserve(p.size());
                        insert(end(), p.begin(), p.end());
                        white   = p.white  ;
                        black   = p.black  ;
                        index   = p.index  ;
                        first   = p.first  ;
                        last    = p.last   ;
                        name    = p.name   ;
                        p_space = p.p_space;
                        return *this;
                    }
    //------------------------------------------------------------------------
    LaserBoy_palette&   operator += (const LaserBoy_palette& p)
                    {
                        reserve(size() + p.size());
                        insert(end(), p.begin(), p.end());
                        return *this;
                    }
    //------------------------------------------------------------------------
    LaserBoy_color& operator [] (const int& index)
                {
                    if(index >= 0 && index < number_of_colors())
                        return at(index);
                    return dud;
                }
    //------------------------------------------------------------------------
    int    number_of_colors () const {   return (int)size();   }
    //------------------------------------------------------------------------
    void   set_palette_item ( const int&    color_index,
                              const u_char& r,
                              const u_char& g,
                              const u_char& b
                            )
            {
                if(color_index >= 0 && color_index < number_of_colors())
                {
                    at(color_index).r = r;
                    at(color_index).g = g;
                    at(color_index).b = b;
                }
            }
    //------------------------------------------------------------------------
    void   wipe()
            {
                for(int i = 0; i < number_of_colors(); i++)
                    at(i) = (u_char)0x00;
                return;
            }
    //------------------------------------------------------------------------
    LaserBoy_palette&   reorder();
    //------------------------------------------------------------------------
    int    from_ifstream_ild (ifstream& in,  const LaserBoy_header& header)       ;
    void   to_ofstream_ild   (ofstream& out,       LaserBoy_header& header) const ;
    //------------------------------------------------------------------------
    int    from_ascii_file       (string file)                ;
    void   save_as_ascii         (string file)          const ;
    //------------------------------------------------------------------------
    int    to_bmp_palette   (struct LaserBoy_bmp* bmp ) const ;
    //------------------------------------------------------------------------
    void   find_factors          (                         )      ;
    int    best_match            (LaserBoy_color rgb       )      ;
    void   best_reduction        (                         )      ;
    void   shade                 (u_char shade             )      ;
    void   tint                  (u_char tint              )      ;
    bool   unite                 (const LaserBoy_palette& palette);
    //------------------------------------------------------------------------
    bool             in_use;

    u_int            white,
                     black;

    int              index,
                     first,
                     last;

    LaserBoy_color   dud;

    string           name;

    LaserBoy_space  *p_space;
};

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
