//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
// Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009 James Lehman.
// This source is distributed under the terms of the GNU General Public License.
//
// LaserBoy_space.hpp is part of LaserBoy.
//
// LaserBoy is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// LaserBoy is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LaserBoy. If not, see <http://www.gnu.org/licenses/>.
//
//############################################################################
#ifndef __LASERBOY_SPACE_DEFINITIONS__
#define __LASERBOY_SPACE_DEFINITIONS__

//############################################################################
#include "LaserBoy_frame_set.hpp"
#include "LaserBoy_palette_set.hpp"
//#include "LaserBoy_GUI_base.hpp"

class LaserBoy_GUI_base;

//############################################################################
class LaserBoy_space : public LaserBoy_frame_set, public LaserBoy_palette_set
{
public:
    LaserBoy_space(LaserBoy_GUI_base* gui);
    //------------------------------------------------------------------------
virtual
   ~LaserBoy_space()
    {
        bmp_free (&bmp);
        bmp_free (&background);
    }
    //------------------------------------------------------------------------
    LaserBoy_frame& operator [] (int index) // does not set frame_index
                    {
                        if(number_of_frames())
                        {
                            if(index >= number_of_frames())
                                index = 0;
                            if(index < 0)
                                index = number_of_frames() - 1;
                            return LaserBoy_frame_set::at(index);
                        }
                        return null_frame;
                    }
    //------------------------------------------------------------------------
    LaserBoy_space& operator += (const LaserBoy_frame_set& frames)
                    {
                        LaserBoy_frame_set::insert(LaserBoy_frame_set::end(),
                                                   frames.begin(), frames.end()
                                                  );
                        renumber();
                        return *this;
                    }
    //------------------------------------------------------------------------
    LaserBoy_space& operator  = (const LaserBoy_space& space)
                    {
                        LaserBoy_palette_set::clear();
                        LaserBoy_palette_set::reserve(space.number_of_palettes());

                        LaserBoy_palette_set::insert (LaserBoy_palette_set::end(),
                                                      space.LaserBoy_palette_set::begin(),
                                                      space.LaserBoy_palette_set::end()
                                                     );
                        //....................................................
                        LaserBoy_frame_set::clear();
                        LaserBoy_frame_set::reserve(space.number_of_frames());
                        LaserBoy_frame_set::insert (LaserBoy_frame_set::end(),
                                                    space.LaserBoy_frame_set::begin(),
                                                    space.LaserBoy_frame_set::end()
                                                   );
                        renumber();
                        return *this;
                    }
    //------------------------------------------------------------------------
    LaserBoy_space& operator  = (const LaserBoy_frame_set& frames)
                    {
                        LaserBoy_frame_set::clear();
                        LaserBoy_frame_set::reserve(frames.number_of_frames());
                        LaserBoy_frame_set::insert (LaserBoy_frame_set::end(),
                                                    frames.LaserBoy_frame_set::begin(),
                                                    frames.LaserBoy_frame_set::end()
                                                   );
                        renumber();
                        return *this;
                    }
    //------------------------------------------------------------------------
    LaserBoy_space& operator += (const LaserBoy_frame& frame)
                    {
                        LaserBoy_frame_set::push_back(frame);
                        renumber();
                        return *this;
                    }
    //------------------------------------------------------------------------
    LaserBoy_space& operator += (const LaserBoy_palette_set& palettes)
                    {
                        LaserBoy_palette_set::insert(LaserBoy_palette_set::end(),
                                                     palettes.begin(),
                                                     palettes.end()
                                                    );
                        return *this;
                    }
    //------------------------------------------------------------------------
    LaserBoy_space& operator += (const LaserBoy_palette& palette)
                    {
                        LaserBoy_palette_set::push_back(palette);
                        return *this;
                    }
    //------------------------------------------------------------------------
    void     clear_frames()
                    {
                        LaserBoy_frame_set::clear();
                        return;
                    }
    //------------------------------------------------------------------------
    void     clear_palettes()
                    {
                        LaserBoy_palette_set::clear();
                        return;
                    }
    //------------------------------------------------------------------------
    void     clear_unused_palettes        ();
    void     sync_rgb_and_palette         ();
    void     best_reduce_all_to_palette   ();
    //------------------------------------------------------------------------
    void     load_cfg_file                ();
    void     save_cfg_file                ();
    //------------------------------------------------------------------------
    void     load_frame_effects           ();
    void     load_frame_set_effects       ();
    //------------------------------------------------------------------------
    bool     call_frame_effect            (string name);
    bool     call_frame_set_effect        (string name);
    //------------------------------------------------------------------------
    void     load_background_bitmap       (struct LaserBoy_bmp* bmp);
    bool     load_background_bitmap       (string file             );
    void     clear_background_bitmap      ();
    void     recolor_background           ();
    //------------------------------------------------------------------------
    void     simplify_view_angle()
                {
                    while(view_angle.x < 0)
                        view_angle.x += two_pi;
                    while(view_angle.x > two_pi)
                        view_angle.x -= two_pi;

                    while(view_angle.y < 0)
                        view_angle.y += two_pi;
                    while(view_angle.y > two_pi)
                        view_angle.y -= two_pi;

                    while(view_angle.z < 0)
                        view_angle.z += two_pi;
                    while(view_angle.z > two_pi)
                        view_angle.z -= two_pi;
                }
    //------------------------------------------------------------------------
    LaserBoy_bmp* p_bmp           () { return &bmp;            }
    LaserBoy_bmp* p_background    () { return &background;     }
    LaserBoy_bmp* p_background_bmp() { return &background_bmp; }
    //------------------------------------------------------------------------
    void undo_wave_temp(ifstream &in, ofstream &out, string file_name)
    {
        if(in.is_open())
            in.close();
        if(out.is_open())
            out.close();
        remove(file_name.c_str());
        rename(  (file_name + ".tmp").c_str()
               , file_name.c_str()
              );
        return;
    }
    //------------------------------------------------------------------------
    bool    apply_wave_offsets_prep       (string file);
    bool    overwirte_wave_offsets_prep   (string file);
    bool    invert_wave_prep              (string file);
    bool    invert_signals_prep           (string file);
    bool    clear_wave_polarity_list_prep (string file);
    //------------------------------------------------------------------------
    bool    split_wave_XY_r_g_b_i_LR_prep (   string file
                                            , string dir_in
                                            , string dir_laser
                                            , string dir_audio
                                          );
    //------------------------------------------------------------------------
    bool    join_XY_r_g_b_i_waves_prep    (   string file_XY
                                            , string file_r
                                            , string file_g
                                            , string file_b
                                            , string file_i
                                            , string file_out
                                          );
    //------------------------------------------------------------------------
    bool    join_XY_rg_b_waves_prep       (   string file_XY
                                            , string file_rg
                                            , string file_b
                                            , string file_out
                                          );
    //------------------------------------------------------------------------
    bool    join_XY_i_LR_waves_prep       (   string file_XY
                                            , string file_i
                                            , string file_LR
                                            , string file_out
                                          );
    //------------------------------------------------------------------------
    bool    add_audio_to_wave_prep        (   string file_laser
                                            , string file_audio
                                            , string file_out
                                          );
    //------------------------------------------------------------------------
    void apply_wave_offsets(ifstream&            wave_ifstream,
                            ofstream&            wave_ofstream,
                            LaserBoy_wave_header wave_in_header
                           );
    //------------------------------------------------------------------------
    void overwirte_wave_offsets(ifstream&            wave_ifstream,
                                ofstream&            wave_ofstream,
                                LaserBoy_wave_header wave_in_header
                               );
    //------------------------------------------------------------------------
    void invert_wave(ifstream&            wave_ifstream,
                     ofstream&            wave_ofstream,
                     LaserBoy_wave_header wave_in_header
                    );
    //------------------------------------------------------------------------
    void invert_signals(ifstream&            wave_ifstream,
                        ofstream&            wave_ofstream,
                        LaserBoy_wave_header wave_in_header
                       );
    //------------------------------------------------------------------------
    void clear_wave_polarity_list(ifstream&            wave_ifstream,
                                  ofstream&            wave_ofstream,
                                  LaserBoy_wave_header wave_in_header
                                 );
    //------------------------------------------------------------------------
    void split_wave_XY_r_g_b_i_LR(LaserBoy_wave_header wave_in_header,
                                  ifstream&            wave_ifstream,
                                  ofstream&            wave_ofstream_XY,
                                  ofstream&            wave_ofstream_r ,
                                  ofstream&            wave_ofstream_g ,
                                  ofstream&            wave_ofstream_b ,
                                  ofstream&            wave_ofstream_i ,
                                  ofstream&            wave_ofstream_LR
                                 );
    //------------------------------------------------------------------------
    void join_XY_r_g_b_i_waves(ifstream&     in_XY,
                               ifstream&     in_r,
                               ifstream&     in_g,
                               ifstream&     in_b,
                               ifstream&     in_i,
                               ofstream&     out
                              );
    //------------------------------------------------------------------------
    void join_XY_rg_b_waves(ifstream&     in_XY,
                            ifstream&     in_rg,
                            ifstream&     in_b,
                            ofstream&     out
                           );
    //------------------------------------------------------------------------
    void join_XY_i_LR_waves(ifstream&     in_XY,
                            ifstream&     in_i ,
                            ifstream&     in_LR,
                            ofstream&     out
                           );
    //------------------------------------------------------------------------
    void add_audio_to_wave( LaserBoy_wave_header wave_in_header_laser,
                            LaserBoy_wave_header wave_in_header_audio,
                            ifstream&            in_laser,
                            ifstream&            in_audio,
                            ofstream&            out
                          );
    //------------------------------------------------------------------------
    void add_silence_to_wave( LaserBoy_wave_header wave_in_header_laser,
                              ifstream&            in_laser,
                              ofstream&            out
                            );
    //------------------------------------------------------------------------
    string LaserBoy_wave_signal_id_to_name(const short& signal_id) const;
    string LaserBoy_wave_LSB_tag_to_name  (const short& LSB_tag  ) const;
    void   tell(LaserBoy_wave_header header, string label) const;
    //------------------------------------------------------------------------
    bool    show_vertices           ,
            show_blanking           ,
            show_cursors            ,
            show_fixed_bounds       ,
            show_fixed_origin       ,
            show_floating_axis      ,
            show_floating_bounds    ,
            show_fulcrum            ,
            show_effects_generation ,
            show_background_bitmap  ,
            show_menu               ,
            show_palettes           ,
            show_stats              ,
            show_TUI_clue           ,
            show_frame_stats        ,
            show_frame_set_stats    ,
            ui_visuals_flip_flop    ,
            loop_animation          ,
            approximate_frame_rate  ,
            wave_show_vertices      ,
            wave_show_blanks        ,
            wave_real_time          ,
            wave_show_inverted      ,
            wave_clean_screen       ,
            wave_loop               ,
            auto_scale_dxf          ,
            maintain_dxf_origin     ,
            black_dwell_vertices    ,
            auto_flatten_z          ,
            no_equivalent_vectors   ,
            no_dots                 ,
            invert_wave_output      ,
            auto_optimize           ,
            no_black_match          ,
            allow_lit_black         ,
            dumb_file_format        ;

    u_char  black_threshold         ;

    int     sample_rate             ,
            lit_dwell_overhang      ,
            dwell_on_dot            ,
            output_bmp_size         ,
            font_size_factor        ,
            signal_if_not_z         ,
            unframed_wave_stride    ,
            displacement_step       ,
            signal_bit_mask      [8],
            signal_polarity      [8],
            wave_offsets         [8];

    double  lit_delta_max           ,
            blank_delta_max         ,
            max_dwell_microsec      ,
            insignificant_distance  ,
            insignificant_angle     ,
            recolor_span_factor     ,
            rendered_arc_angle      , // in degrees
            frames_per_second       ,
            still_frame_duration    ,
            rotation_step           , // in radians!
            magnitude_step          ,
            magnitude_step_up       ,
            magnitude_step_dn       ;

    struct
    LaserBoy_bmp       bmp           , // a memory clone of screen
                       background    , // empty for clearing screen
                       background_bmp; // possibly a bitmap picture

    string             bg_file_name  ;

    LaserBoy_color     bg_color      ;

    LaserBoy_3D_double fulcrum       ,
                       view_angle    ,
                       view_scale    ,
                       view_offset   ;

    vector<string>                     frame_effect_names     ,
                                       frame_set_effect_names ;
    vector<LaserBoy_frame_effect>      frame_effects          ;
    vector<LaserBoy_frame_set_effect>  frame_set_effects      ;

    LaserBoy_space*     p_space;
    LaserBoy_GUI_base*  GUI;
    //------------------------------------------------------------------------
};

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
