//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
// Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009 James Lehman.
// This source is distributed under the terms of the GNU General Public License.
//
// LaserBoy_wave.hpp is part of LaserBoy.
//
// LaserBoy is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// LaserBoy is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LaserBoy. If not, see <http://www.gnu.org/licenses/>.
//
//############################################################################
#ifndef __LASERBOY_WAVE_DEFINITIONS__
#define __LASERBOY_WAVE_DEFINITIONS__

//############################################################################
#include "LaserBoy_common.hpp"

//############################################################################
class LaserBoy_wave_optimization_parameters
{
public:
    LaserBoy_wave_optimization_parameters()
              : lit_dwell_overhang     (-1), // -1 indicates undefined
                lit_delta_max          (-1),
                blank_delta_max        (-1),
                max_dwell_microsec     (-1),
                insignificant_distance (-1),
                insignificant_angle    (-1),
                frames_per_second      (-1)
                                                      {}
    //------------------------------------------------------------------------
   ~LaserBoy_wave_optimization_parameters()           {}
    //------------------------------------------------------------------------
    int    lit_dwell_overhang;
    float  lit_delta_max,
           blank_delta_max,
           max_dwell_microsec,
           insignificant_distance,
           insignificant_angle,
           frames_per_second;
};

//############################################################################
class LaserBoy_wave_header
{
public:
    LaserBoy_wave_header(int r, u_int m, short c)
                  : num_samples        (0),
                    sample_rate        (r),
                    LaserBoy_wave_mode (m),
                    num_channels       (c),
                    bits_per_sample    (16),
                    version            (LASERBOY_WAVE_VERSION),
                    parms              ()
                                                      {}
    //------------------------------------------------------------------------
    LaserBoy_wave_header    (const LaserBoy_wave_header& header     );
    LaserBoy_wave_header    (const LaserBoy_wave_header& header, int); // copy 8ch from 6ch
    LaserBoy_wave_header    (ifstream& in                           );
    //------------------------------------------------------------------------
   ~LaserBoy_wave_header    ()                        {}
    //------------------------------------------------------------------------
    void     to_ofstream_wave (ofstream& out) const;
    //------------------------------------------------------------------------
    int      num_samples    ,
             sample_rate    ,
             offset      [8];
    //------------------------------------------------------------------------
    u_char   resolution  [8];
    //------------------------------------------------------------------------
    u_int    LaserBoy_wave_mode; // up to 32 binary flags
    //------------------------------------------------------------------------
    short    num_channels   ,
             bits_per_sample,
             signal_id   [8],
             LSB_tag     [8];
    //------------------------------------------------------------------------
    string   version        ;
    //------------------------------------------------------------------------
    LaserBoy_wave_optimization_parameters parms;
};

//############################################################################
class LaserBoy_wave_sample
{
public:
    LaserBoy_wave_sample(int n)
        {
            num_channels = n;
            channel = new short[num_channels];
            memset((void*)channel, 0x00, num_channels * sizeof(short));
        }
    //------------------------------------------------------------------------
   ~LaserBoy_wave_sample()    {    delete channel;    }
    //------------------------------------------------------------------------
    bool from_ifstream_wave(ifstream& in)
            {
                if(!in.eof())
                {
                    in.read((char*)channel, num_channels * sizeof(signed short));
                    return true;
                }
                return false; // returns false if EOF
            }
    //------------------------------------------------------------------------
    bool from_ifstream_wave_inverted(ifstream& in)
            {
                if(!in.eof())
                {
                    in.read((char*)channel, num_channels * sizeof(signed short));
                    negate();
                    return true;
                }
                return false; // returns false if EOF
            }
    //------------------------------------------------------------------------
    bool from_ifstream_wave_polarity(ifstream& in, int polarity[8])
            {
                if(!in.eof())
                {
                    in.read((char*)channel, num_channels * sizeof(signed short));
                    for(int i = 0; i < num_channels; i++)
                        if(polarity[i] == 1)
                        {
                            if(channel[i] >= LASERBOY_MIN_SHORT)
                                channel[i] = -channel[i];
                            else
                                channel[i] = LASERBOY_MAX_SHORT;
                        }
                    return true;
                }
                return false; // returns false if EOF
            }
    //------------------------------------------------------------------------
    void to_ofstream_wave(ofstream& out, LaserBoy_wave_header& header)
            {
                out.write((char*)channel, num_channels * sizeof(signed short));
                header.num_samples++;
                return;
            }
    //------------------------------------------------------------------------
    void to_ofstream_wave_inverted(ofstream& out, LaserBoy_wave_header& header)
            {
                negate();
                out.write((char*)channel, num_channels * sizeof(signed short));
                negate();
                header.num_samples++;
                return;
            }
    //------------------------------------------------------------------------
    LaserBoy_wave_sample& operator = (const LaserBoy_wave_sample& sample)
                    {
                        delete channel;
                        num_channels = sample.num_channels;
                        channel = new short[num_channels];
                        memcpy((void*)channel, (void*)sample.channel, num_channels * sizeof(short));
                        return *this;
                    }
    //------------------------------------------------------------------------
    void negate()
            {
                for(int i = 0; i < num_channels; i++)
                    channel[i] = -channel[i];
                return;
            }
    //------------------------------------------------------------------------
    LaserBoy_wave_sample  operator + (const LaserBoy_wave_sample& s) // side-by-side as tracks
                    {
                        LaserBoy_wave_sample sum(num_channels + s.num_channels);
                        memcpy(   (void*)sum.channel,
                                  (void*)channel,
                                  num_channels * sizeof(short)
                              );
                        memcpy(   (void*)(sum.channel + num_channels), // + offset for above
                                  (void*)s.channel,
                                  s.num_channels * sizeof(short)
                              );
                        return sum;
                    }
    //------------------------------------------------------------------------
    signed short   *channel     ;
    int             num_channels;
};

//############################################################################
typedef LaserBoy_wave_sample*  LaserBoy_wave_sample_pointer;

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
